/******************************************************************************
 * MseedFileName() : Function to compute a filename from station data, time   *
 *                   and type of miniseed archive                             *
 ******************************************************************************/
// Defines
#define FILE_SEPARATOR "/" // File separator

#include <earthworm.h>
#include <libmseed2.h>
#include <math.h>

// Local Includes
#include "msdb.h"
#include "mseedfilename.h"

static void setEndData(MSDATA *msdata) {
  msdata->datasamples = NULL;
  msdata->starttime = 0.;
  msdata->endtime = 0.;
}

/**
 * Clear the miniSEED data.
 * @param msdata the miniSEED data
 */
void clearMsdata(MSDATA *msdata) { memset(msdata, 0, sizeof(MSDATA)); }

/**
 * Free the miniSEED data.
 * @param msdata the miniSEED data
 */
int freeMsdata(MSDATA *msdata) {
  int retcode =
      ms_readmsr_r(&msdata->msfp, &msdata->msr, NULL, 0, NULL, NULL, 0, 0, 0);
  msdata->msfp = NULL;
  msdata->msr = NULL;
  return retcode;
}

/**
 * Get the delta.
 * @param samprate the sample rate
 * @return the delta
 */
double getDelta(double samprate) { return pow(samprate, -1); }

static double getEndtime(MSDATA *msdata) {
  return msdata->starttime +
         (msdata->msr->samplecnt - 1) / msdata->msr->samprate;
}

static double getStarttime(MSDATA *msdata) {
  return msr_starttime(msdata->msr) / 1000000.0;
}

/**
 * Initialize the miniSEED data after reading the data.
 * @param msdata the miniSEED data
 */
void initMsdata(MSDATA *msdata) {
  msdata->samprate = msdata->msr->samprate;
  msdata->delta = getDelta(msdata->samprate);
  msdata->datasamples = (int *)msdata->msr->datasamples;
  msdata->starttime = getStarttime(msdata);
  msdata->endtime = getEndtime(msdata);
}

/**
 * Determines if the end of data.
 * @return true if the end of data, false othewise.
 */
int isEndData(MSDATA *msdata) { return msdata->datasamples == NULL; }

/**
 *Get the miniSEED file name.
 *
 *Input:
 * @param mseeddir is the miniSEED directory
 * @param t is the Unix epoch time in seconds
 * @param stat is the station code
 * @param chan is the channel code
 * @param net is the network code
 * @param loc is the location code
 * @param archiveFormat is the archive format (FORMAT_BUD, FORMAT_SCP)
 *
 *Output:
 *@param mseedfile is the miniSEED file
 *@return the miniSEED file name.
 */
char *MseedFileName(char *mseedfile, char *mseeddir, double t, char *stat,
                    char *chan, char *net, char *loc, int archiveFormat) {
  BTime btime;
  *mseedfile = '\0';
  // Compute btime to have a day of year estimate
  if ((ms_hptime2btime((hptime_t)(t * 1000000 + 0.5), &btime)) == -1) {
    logit("e", "Error computing time: %f\n", t);
    exit(-1);
  }
  if (strcmp(loc, "--") == 0) {
    loc = "";
  }
  // Compute file name based on archive type
  switch (archiveFormat) {
  case FORMAT_BUD:
    snprintf(mseedfile, MSEEDFILE_SZ, "%s/%s/%s/%s.%s.%s.%s.%d.%03d", mseeddir,
             net, stat, stat, net, loc, chan, btime.year, btime.day);
    break;
  case FORMAT_SCP:
    snprintf(mseedfile, MSEEDFILE_SZ, "%s/%d/%s/%s/%s.D/%s.%s.%s.%s.D.%d.%03d",
             mseeddir, btime.year, net, stat, chan, net, stat, loc, chan,
             btime.year, btime.day);
    break;
  case FORMAT_MSEEDINDEX:
    msdb_get_filename(mseedfile, mseeddir, t, stat, chan, net, loc);
    break;
  default:
    logit("e", "invalid archive format: %d\n", archiveFormat);
    exit(-1);
  }
  return mseedfile;
}

/**
 * Read the miniSEED data.
 * @param msdata the miniSEED data
 * @return MS_ENDOFFILE if end of file, MS_NOERROR  if no error, non-zero if
 * error.
 */
int readData(MSDATA *msdata) {
  int retcode;
  retcode = ms_readmsr_r(&msdata->msfp, &msdata->msr, msdata->mseedfile, 0,
                         NULL, NULL, 1, 1, 0);
  if (retcode == MS_NOERROR) {
    initMsdata(msdata);
    if (msdata->debug > 1) {
      logit("o",
            "readData start=%f, end=%f, samplecnt=%d, samprate=%f, delta=%f\n",
            msdata->starttime, msdata->endtime, (int)msdata->msr->samplecnt,
            msdata->samprate, msdata->delta);
    }
  } else {
    if (retcode != MS_ENDOFFILE) {
      logit("e", "Cannot read %s: %s\n", msdata->mseedfile,
            ms_errorstr(retcode));
    }
    setEndData(msdata);
  }
  return retcode;
}
