
       /********************************************************
        *                      port_nt.c                       *
        *                                                      *
        *  Get Trimble GPS data through a com port.            *
        *  GPS data is assumed to be in TSIP format.           *
        *  The functions in this file are Windows compatible.  *
        ********************************************************/


#include <windows.h>
#include <stdlib.h>
#include <time.h>
#include <earthworm.h>      // for logit
#include "adsendxs.h"

static HANDLE handle = 0;


/************************************************
  LogLastSystemError

  Retrieve and log the last system error message.
 ************************************************/

void LogLastSystemError( void )
{
   LPVOID lpMsgBuf;
   DWORD dw = GetLastError();

   FormatMessage
   (
      FORMAT_MESSAGE_ALLOCATE_BUFFER |
      FORMAT_MESSAGE_FROM_SYSTEM |
      FORMAT_MESSAGE_IGNORE_INSERTS,
      NULL, dw,
      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
      (LPTSTR) &lpMsgBuf,
      0, NULL
    );

    logit( "et", "%s", lpMsgBuf );
    LocalFree( lpMsgBuf );
    return;
}


/*****************************************************
  OpenPort

  Open the com port and set attributes.
 *****************************************************/

int OpenPort( char PortName[], int BaudRate, int ReadFileTimeout )
{
   COMMTIMEOUTS timeouts = {0};
   DCB          dcb;                    // Data control block

/* Sanity check
   ************/
   switch ( BaudRate )
   {
      case   1200: break;
      case   2400: break;
      case   4800: break;
      case   9600: break;
      case  14400: break;
      case  19200: break;
      case  38400: break;
      case  57600: break;
      case 115200: break;
      case 128000: break;
      case 256000: break;
      default:
         logit( "et", "Invalid baud rate: %d\n", BaudRate );
         return -1;
   }

/* Open the com port for asynchronous input/output.
   Share mode is zero, which allows exclusive access only.
   ******************************************************/
   handle = CreateFile( PortName,                       // COM1 or COM2, ...
                        GENERIC_READ|GENERIC_WRITE,     // Desired access
                        0,                              // Share mode
                        NULL,                           // Default security attributes
                        OPEN_EXISTING,                  // Create parameter
                        FILE_ATTRIBUTE_NORMAL,          // Flags and attributes
                        NULL );                         // Template file

   if ( handle == INVALID_HANDLE_VALUE )
   {
      LogLastSystemError();
      logit( "et", "Error opening port %s\n", PortName );
      return -1;
   }

/* Set PC serial port parameters to the default values
   for a Trimble Thunderbolt E GPS receiver.
   ***************************************************/
   if ( GetCommState(handle, &dcb) == 0 )
   {
      logit( "et", "Error getting state of comm port.\n" );
      LogLastSystemError();
      return -1;
   }

   dcb.BaudRate        = (DWORD)BaudRate;
   dcb.ByteSize        = 8;            // Hardwired
   dcb.Parity          = NOPARITY;     // Hardwired
   dcb.StopBits        = ONESTOPBIT;   // Hardwired
   dcb.fOutxCtsFlow    = FALSE;        // Disable RTS/CTS flow control
   dcb.fOutxDsrFlow    = FALSE;        // Disable DTR/DSR flow control
   dcb.fDsrSensitivity = FALSE;        // Driver ignores DSR
   dcb.fOutX           = FALSE;        // Disable XON/XOFF out flow control
   dcb.fInX            = FALSE;        // Disable XON/XOFF in flow control
   dcb.fNull           = FALSE;        // Don't discard null bytes
   dcb.fAbortOnError   = FALSE;        // Don't abort rd/wr on error

   if ( SetCommState(handle, &dcb) == 0 )
   {
      LogLastSystemError();
      logit( "et", "Error setting comm port state.\n" );
      return -1;
   }

/* Set read and write timeouts on the com port.
   If there any bytes in the input buffer, ReadFile returns
      immediately with the bytes in the buffer.
   If there are no bytes in the input buffer, ReadFile waits
      until a byte arrives and then returns immediately.
   If no bytes arrive within ReadTotalTimeoutConstant
      milliseconds, ReadFile times out.
   Timeouts are not used for write operations.
   *********************************************************/
   timeouts.ReadIntervalTimeout         = MAXDWORD;
   timeouts.ReadTotalTimeoutMultiplier  = MAXDWORD;
   timeouts.ReadTotalTimeoutConstant    = ReadFileTimeout;   // Milliseconds
   timeouts.WriteTotalTimeoutMultiplier = 0;
   timeouts.WriteTotalTimeoutConstant   = 0;

   if ( SetCommTimeouts( handle, &timeouts ) == 0 )
   {
      LogLastSystemError();
      logit( "et", "Error setting serial port timeouts.\n" );
      return -1;
   }
   return 0;
}


/************************************************
  ClosePort

  Close the com port.
 ************************************************/

void ClosePort( void )
{
   if ( handle != 0 )
      CloseHandle( handle );
   return;
}


/******************************************************
  PurgeComInput

  Purge the input com port buffer of any leftover
  characters from the last read.
 ******************************************************/
int PurgeComInput( void )
{
   if ( PurgeComm(handle, PURGE_RXCLEAR) == 0 )
   {
      LogLastSystemError();
      logit( "et", "Error purging input serial port buffer.\n" );
      return -1;
   }
   return 0;
}


/********************************************************
  ReadGpsPacket

  Read one packet from the GPS receiver via com port.
  The com port is left open on return.
  Returns number of bytes read (could be 0).
  Returns -1 if an error occurred.
 ********************************************************/

int ReadGpsPacket( uint8_t packet[] )
{
   uint8_t chr;
   int     npacketchar = 0;
   int     nDLE = 0;
   int     first = 1;

/* Read a packet from the GPS receiver, one byte at a time.
   The TSIP packet format is:
   <DLE><id><data string bytes><DLE><ETX>
   *******************************************************/
   while ( 1 )
   {
      DWORD numRead = 0;
      uint8_t buf[2] = {0};

      if ( ReadFile( handle, buf, 1, &numRead, NULL ) == FALSE )
      {
         LogLastSystemError();
         logit( "et", "ReadFile error in ReadGpsPacket.\n" );
         return -1;
      }

/* Did we get one byte from the GPS, as requested?
   **********************************************/
      if ( numRead == 0 )
      {
         logit( "et", "Error in ReadGpsPacket.  Timed out reading from GPS receiver.\n" );
         return -1;
      }

/* The first character read must be a DLE
   **************************************/
      chr = buf[0];
      if ( first && (chr != DLE) )
      {
         logit( "et", "ReadGpsPacket error.  First character in packet is not DLE.\n" );
         return -1;
      }
      else
         first = 0;

/* Count the total DLE's received
   ******************************/
      if ( chr == DLE ) nDLE++;
      if ( (chr == ETX) && (nDLE % 2 == 0) ) break;

/* Check for buffer overflow
   *************************/
      if ( npacketchar + 1 > MAXGPSPACKETSIZE )
      {
         logit( "et", "ReadGpsPacket buffer overflow.\n" );
         return -1;
      }

/* Save the byte
   *************/
      packet[npacketchar++] = chr;
   }
   packet[npacketchar++] = chr;   // Save the final ETX character
   return npacketchar;            // Return number of bytes read
}


/************************************************
  WriteGpsPacket

  Write packet to GPS receiver via com port.
  Returns  0 if all ok,
          -1 if a write error occurred
 ************************************************/

int WriteGpsPacket( uint8_t packet[], int nchar )
{
   DWORD numWritten;

   if ( WriteFile( handle, packet, nchar, &numWritten, 0 ) == 0 )
   {
      LogLastSystemError();
      logit( "et", "Error writing GPS packet.\n" );
      return -1;
   }

   if ( (int)numWritten < nchar )
   {
      logit( "et", "Requested write of %d characters to GPS.  Only %u were written.\n",
              nchar, numWritten );
      return -1;
   }
   return 0;
}
