/*****************************************************************/
/*                                                               */
/* Name:  "latency.c"                                            */
/* Author: Andrew Good                                           */
/*                                                               */
/* Description:                                                  */
/*                                                               */
/*  Program to read waveform messages from the WAVE_RING,        */
/*  calculate latency values and write them to a LATENCY_RING.   */
/*                                                               */
/*                                                               */
/*  Three modes of operation are implemented:                    */
/*                                                               */
/*		1) When a message is detected on the Wave Ring, calculate  */
/*       the latency and output it immediately in real time.     */
/*                                                               */
/*    2) Output in real time, but force an output sample rate    */
/*       of 1 record per second.                                 */
/*                                                               */
/*    3) Use linear interpolation to calculate latency values    */
/*       to attach to each record and archive them at an         */
/*       artificially-induced rate of 1 sample per second        */
/*       (this requires storing enough data values to do the     */
/*       interpolations, precluding real-time output.            */
/*                                                               */
/*****************************************************************/

/*
 *
 * Copyright (c) 2017 California Institute of Technology.
 *
 * All rights reserved.
 *
 * This program is distributed WITHOUT ANY WARRANTY whatsoever.
 *
 * Do not redistribute this program without written permission.
 *
 */

#include <ctype.h>
#include <earthworm.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <swap.h>
#include <time_ew.h>
#include <trace_buf.h>
#include <transport.h>
#include <unistd.h>
#include "latency_scnl.h"
#include "net.h"
#include <kom.h>

#define TYPE_NOTUSED 254
#define NLOGO 1 // Only one input datatype is expected (TraceBuf2)
/* Cast quality fields to unsigned char when you want convert to hexadecimal. */
#define CAST_QUALITY(q) ((unsigned char)q)

#define RAW_MODE 0
#define FORCED_MODE 1
#define INTERP_MODE 2
#define SNW_MODE 3

/* versioning introduced by ISTI when incorporated into Earthworm repo */
#define VERSION "1.2.0 - 2021-02-16"

static int pid;                         // Process ID
static char ProgName[100];              /* program name for logging purposes  */
static unsigned char Type_Heartbeat;    // Type for Earthworm heartbeat
static unsigned char Type_Error;        // Type for Earthworm error
static unsigned char OutputType = '\0'; // Output type for this module
static int debug = 0;                   // Toggle debug messages
static int testmode = 0;      // Test mode on/off (config file parameter)
static int renameChannel = 1; // Apply channel renaming conventions
static int slept = 0;         // 1 if a sleep has just occurred, 0 otherwise
static int typeLatencyScnl = 0; // 0 for TYPE_TRACEBUF2, 1 for TYPE_LATENCY_SCNL, 2 for TYPE_LATENCY_SCNL_TEXT, 3 for TYPE_LATENCY_SCNL_BIN
static int ValidTraceBuf2 = 0;

static int sock; // Socket

static double **inputTime; // Array for storing input times
static double **inputLat;  // Array for storing input latencies
static double
    *refTime; // Time (second) to compute  output latency value for
              // (used in Forced case, corresponds to nominal sample endTime)

/////////////////////////////////////////////////////////////////////////
static double
    *scnl_lastPacketTime;    // Corresponds to scnl_time[i].last_packet_time in
                             // sniffwave
static double **feedLat;     // Array for storing feed latencies
static double **packet_size; // Packet sizes for this channel
/////////////////////////////////////////////////////////////////////////

static int *nSamp; // Count of records read in for each SCNL tag
static int
    *sleptStatus; // For each input, track  whether it has returned from a sleep

static int nTag; // Counter for number of times a particular tag
                 // has been encountered

// SCNL tags that we want to process
struct TAG {
  char sta[TRACE_STA_LEN];
  char cha[TRACE_CHAN_LEN];
  char net[TRACE_NET_LEN];
  char loc[TRACE_LOC_LEN];
  int mode;
};

// Instance of TAG structure that we will be using
static struct TAG *tag;

static SHM_INFO region_in; // Shared memory information about input (Wave) ring
static SHM_INFO
    region_out; // Shared memory information about output (Latency) ring

// Ring messages are:  "Logo + message".  In the case of a TraceBuf2 message,
// the "message" is "header + data".

static MSG_LOGO getLogo; // Message description (discovered by analysis)
                         // Note: changes getLogo[NLOGO] to getLogo, and
                         // reference below to &getLogo.

static MSG_LOGO inputLogo, outputLogo; // Message descriptions ('transport.h')

static long inputKey;     // Key to the transport ring we read from
static long outputKey;    // Key to the transport ring we write to
static long inputLength;  // Size that input  message is found to be
static long outputLength; // Length of output message

static char inmsg[MAX_TRACEBUF_SIZ]; // Buffer to store messages from Wave Ring
static char
    outmsg[MAX_TRACEBUF_SIZ]; // Buffer to store messages for Latency Ring

static char inRing[1024];   // Name of input ring (e.g. "WAVE_RING")
static char outRing[1024];  // Name of output ring (e.g. "LATENCY_RING")
static char scnlFile[1024]; // Name of SCNL list file
static char hostName[1024]; // Name of host machine
static char sockName[1024]; // Name of client machine
static int portNum;         // Port number

static unsigned char InstId; // Local installation ID (issued)
static unsigned char ModId;  // Module ID (in "earthworm.d", MOD_LATENCY=150)

static TRACE2_HEADER *trh;        // Pointer to the beginning of "inmsg"
static TRACE2_HEADER *trout;      // Pointer to the beginning of "outmsg"
static char orig_datatype[3];

// Pointer(s) to beginning of message data (array address + header size)
// within "inmsg" and "outmsg".  Short (i2) values are not expected, but allowed
// for. Test data is all of type i4.

static short *short_data_in;
static short *short_data_out;
static int *long_data_in;
static int *long_data_out;

static double d_now;    // Current system clock time
static double then;     // Previous heartbeat time
static double timeout;  // Timeout parameter
static double interval; // Integration interval for interpolation calculations
static double
    sleepTime; // Amount of time to sleep when no data is detected on a read
static double increment;     // Time between output records
static double samprate;      // Output sampling rate
static char LocationCode[3]; // location code for output packets

static double t_offset = 0; // Amount to offset input time by when in test mode

// Output packet header values

static char chan0;
static char chan1;
static char v0;
static char v1;
static char datatype[32];
static char pad[32];
static int pin;
static int outNsamp;

// Return index from array scnl_time for SCNL (sta, cha, net, loc).
// If the SCNL is not present yet, it is inserted.
// int search_scnl_time(SCNL_LAST_SCNL_PACKET_TIME *scnl_time, int *n_scnl_time,
// char *sta, char *cha, char *net, char *loc);

// Check if there is a gap from the last packet and the current one in trh.
// A message is printed to stderr if a gap or an overlap occurs.
// int  check_for_gap_overlap_and_output_message(TRACE2_HEADER  *trh,
// SCNL_LAST_SCNL_PACKET_TIME *scnl_time, int *n_scnl_time);

// Compute all packet latency values that it is possible to compute with the
// data that has been read in so far, and output those latencies to the
// LATENCY_RING
static int getRawLatency();
static int getForcedLatency();
static int getInterpolatedLatency();
static int getSNWLatency();
static void status(unsigned char type, short ierr, char *note);

/************************************************************************************/
/* Main method */
/************************************************************************************/

int main(int argc, char **argv) {
  unsigned char InstWildcard;   // We accept messages from any installation
  unsigned char ModWildcard;    // We accept message from any module
  unsigned char Type_TraceBuf2; // Expected input type for this module

  FILE *fin; // Input file pointer

  char line[1024];     // String to hold SCNL tag value
  char modeString[32]; // String to hold the operational mode value
  char *strptr;        // Utility string pointer
  char *begin, *end;   // Utility character pointers
  int i;               // Utility counter
  int rc;              // Return code
  int len;             // Length of a line from SCNL input file

  int nLogo = NLOGO; // Number of logo patterns we are willing to match
  int iTag;          // Index of the tag we are processing (within tag list)

  int defaultMode = RAW_MODE;

  double DL, TL, ML;    // Transport latency, Total latency, Mean latency
  double FL;            // Feed latency
  double heartbeatTime; // Time between heartbeats

  unsigned char sequence_number = 0; // Used by tport_copyfrom
  int flag;                          // transport flag value

  strcpy(hostName, "");
  LocationCode[0] = 0;

  // Check command line arguments
  if (argc < 2) {
    fprintf(stderr, "Error: latency - Insufficient arguments\n");
    fprintf(stderr, "Usage: latency <configFile>\n");
    fprintf(stderr, "Version: %s\n", VERSION);
    return -1;
  } else if (argc > 2) {
    if (strcmp(argv[2], "-d") == 0 || strcmp(argv[2], "-d1") == 0) {
      debug = 1;
    } else if (strcmp(argv[2], "-d2") == 0) {
      debug = 2;
    } else if (strcmp(argv[2], "-d3") == 0) {
      debug = 3;
    } else {
      fprintf(stderr, "Error: Invalid arguments.\n");
      return -1;
    }
  }

  strcpy(ProgName, argv[1]);
  strptr = strchr(ProgName, '.');
  if (strptr)
    *strptr = '\0';

  /* Initialize name of log-file & open it
  ***************************************/
  logit_init(argv[1], 0, 1024, 1);
  logit("t", "latency: version %s\n", VERSION);

  if (debug) {
    logit("", "Debug Level = %d.\n", debug);
  }

  pid = getpid(); // Process ID does not change

  if (debug > 1) {
    logit("", "Process ID = %d\n", pid);
  }

  // Initialize pointers
  trh = (TRACE2_HEADER *)inmsg;
  trout = (TRACE2_HEADER *)outmsg;
  long_data_in = (int *)(inmsg + sizeof(TRACE2_HEADER));
  short_data_in = (short *)(inmsg + sizeof(TRACE2_HEADER));
  long_data_out = (int *)(outmsg + sizeof(TRACE2_HEADER));
  short_data_out = (short *)(outmsg + sizeof(TRACE2_HEADER));

  if (debug > 1) {
    logit("", "Initialized pointers\n");
  }

  // Read configuration file

  if (k_open(argv[1]) == 0) {
    logit("e", "Error: Unable to open configuration file\n");
    return -1;
  }

  strcpy(inRing, "");   // Initialize inRing to empty
  strcpy(outRing, "");  // Initialize outRing to empty
  strcpy(scnlFile, ""); // Initialize scnlFile to empty

  // TODO: Initialize other values to something so you can do error checks if
  // the config file didn't have them specified.

  while (k_rd()) {
    strptr = k_str(); // Get the first token from line

    // Ignore blank lines
    if (!strptr)
      continue;

    // Ignore comments
    if (strptr[0] == '#')
      continue;

    if (k_its("debug")) {
      debug = k_int();
      if (debug) {
        logit("", "Debug Level = %d.\n", debug);
      }
    }
    // Copy name of Wave Ring
    if (k_its("WaveRingName")) {
      strcpy(inRing, k_str());

      if (debug > 1) {
        logit("", "Wave Ring Name: %s\n", inRing);
      }
    }

    // Copy name of Latency Ring
    if (k_its("LatencyRingName")) {
      strcpy(outRing, k_str());

      if (debug > 1) {
        logit("", "Latency Ring Name: %s\n", outRing);
      }
    }

    // Copy name of SCNL list file
    if (k_its("SCNLfile")) {
      strcpy(scnlFile, k_str());

      if (debug > 1) {
        logit("", "SNCL tag list file: %s\n", scnlFile);
      }
    }

    // Get Module ID
    if (k_its("ModuleId")) {
      if ((strptr = k_str())) {
        // If GetModId cannot find module name in config file, then &ModId will
        // become set to -1, and an error will be registered.

        if (GetModId(strptr, &ModId) == -1) {
          return -1;
        }
      }

      if (debug > 1) {
        logit("", "Module ID: %s\n", strptr);
      }
    }

    // Get operational mode
    if (k_its("Mode")) {
      strcpy(modeString, k_str());

      if (strcmp(modeString, "RAW_MODE") == 0)
        defaultMode = RAW_MODE;
      else if (strcmp(modeString, "FORCED_MODE") == 0)
        defaultMode = FORCED_MODE;
      else if (strcmp(modeString, "INTERP_MODE") == 0)
        defaultMode = INTERP_MODE;
      else if (strcmp(modeString, "SNW_MODE") == 0)
        defaultMode = SNW_MODE;

      if (debug > 1) {
        logit("", "Operational Mode: %s\n", modeString);
      }
    }

    // Get timeout value
    if (k_its("Timeout")) {
      timeout = atof(k_str());

      if (timeout <= 0) {
        logit("e", "Error: Invalid timeout in config file.\n");
        return -1;
      }

      if (debug > 1) {
        logit("", "Timeout interval: %f\n", timeout);
      }
    }

    // Get interval value
    if (k_its("Interval")) {
      interval = atof(k_str());

      if (interval <= 0) {
        logit("e", "Error: Invalid interpolation interval in config file.\n");
        return -1;
      }

      if (debug > 1) {
        logit("", "Interpolation interval: %f\n", interval);
      }
    }

    // Get sleepTime value
    if (k_its("SleepTime")) {
      sleepTime = atof(k_str());

      if (sleepTime <= 0) {
        logit("e", "Error: Invalid sleep time in config file.\n");
        return -1;
      }

      if (debug > 1) {
        logit("", "Sleep interval: %f\n", sleepTime);
      }
    }

    // Get time increment value
    if (k_its("Increment")) {
      increment = atof(k_str());

      if (increment <= 0) {
        logit("e", "Error: Invalid time increment value in config file.\n");
        return -1;
      }

      if (debug > 1) {
        logit("", "Inter-sample time increment: %f\n", increment);
      }
    }

    // Get heartbeat interval
    if (k_its("HeartbeatTime")) {
      heartbeatTime = atof(k_str());

      if (heartbeatTime <= 0) {
        logit("e", "Error: Invalid heartbeat value in config file.\n");
        return -1;
      }

      if (debug > 1) {
        logit("", "Heartbeat interval: %f\n", heartbeatTime);
      }
    }

    // Get channel rename flag
    if (k_its("RenameChannels")) {
      renameChannel = atoi(k_str());

      if (renameChannel < 0) {
        logit("e", "Error: Invalid heartbeat value in config file.\n");
        return -1;
      }

      if (debug > 1) {
        logit("", "RenameChannels: %d\n", renameChannel);
      }
    }

    ////////////////////////////////////////////////////////////////////////////////////
    // Get host machine name
    if (k_its("Host")) {
      strcpy(hostName, k_str());

      if (debug > 1) {
        logit("", "Host Machine: %s\n", hostName);
      }
    }

    // Get client machine name
    if (k_its("Client")) {
      strcpy(sockName, k_str());

      if (debug > 1) {
        logit("", "Client Machine: %s\n", sockName);
      }
    }

    // Get port number
    if (k_its("Port")) {
      portNum = atoi(k_str());

      if (portNum <= 0) {
        logit("e", "Error: Invalid port in config file.\n");
        return -1;
      }

      if (debug > 1) {
        logit("", "portNum: %d\n", portNum);
      }
    }
    ////////////////////////////////////////////////////////////////////////////////////

    // Copy output header values
    char tempStr[32];

    if (k_its("Chan_0")) {
      strcpy(tempStr, k_str());
      chan0 = tempStr[0];
    }
    if (k_its("Chan_1")) {
      strcpy(tempStr, k_str());
      chan1 = tempStr[0];
    }
    if (k_its("Version_0")) {
      strcpy(tempStr, k_str());
      v0 = tempStr[0];
    }
    if (k_its("Version_1")) {
      strcpy(tempStr, k_str());
      v1 = tempStr[0];
    }
    if (k_its("Datatype")) {
      strcpy(datatype, k_str());
    }
    if (k_its("LocationCode")) {
      strcpy(tempStr, k_str());
      if (strlen(tempStr) > 2) {
        logit("e",
              "Error: Invalid location code for output in config file, must be "
              "2 chars: %s\n",
              tempStr);
        return -1;
      }
      strcpy(LocationCode, tempStr);
    }
    // if(k_its("Quality"))                 {       strcpy(quality,   k_str());
    // }
    if (k_its("Pad")) {
      strcpy(pad, k_str());
    }

    //      logit("", "Header info: %c %c %c %c %s %s %s \n", chan0, chan1, v0,
    //      v1, datatype, quality, pad);

    // Get output sample rate value
    if (k_its("Samprate")) {
      samprate = atof(k_str());

      if (samprate <= 0) {
        logit("e", "Error: Invalid sample rate for output in config file.\n");
        return -1;
      }

      if (debug > 1) {
        logit("", "samprate: %f\n", samprate);
      }
    }

    // Get output nsamp
    if (k_its("Nsamp")) {
      outNsamp = atoi(k_str());

      if (outNsamp <= 0) {
        logit("e", "Error: Invalid nsamp in config file.\n");
        return -1;
      }

      if (debug > 1) {
        logit("", "nSamp: %d\n", outNsamp);
      }
    }

    // Get output pinno
    if (k_its("Pin")) {
      pin = atoi(k_str());

      if (pin < 0) {
        logit("e", "Error: Invalid pinno in config file.\n");
        return -1;
      }

      if (debug > 1) {
        logit("", "pinno: %d\n", pin);
      }
    }

    if (k_its("OutputMsgName")) {
      if ((strptr = k_str())) {
        if (GetType(strptr, &OutputType) != 0) {
          logit("e",
                "%s: Invalid message type <%s>! Missing from earthworm.d "
                "or earthworm_global.d\n",
                argv[0], strptr);
          return -1;
        }
        if (strcmp("TYPE_LATENCY_SCNL", strptr) == 0) {
          typeLatencyScnl = 1;
        }
        if (strcmp("TYPE_LATENCY_SCNL_TEXT", strptr) == 0) {
          typeLatencyScnl = 2;
        }
        if (strcmp("TYPE_LATENCY_SCNL_BIN", strptr) == 0) {
          typeLatencyScnl = 3;
        }
      }
    }
    
    if (k_its("ValidTraceBuf2")) {
      ValidTraceBuf2 = atoi(k_str());
      if (debug > 1) {
        logit("", "ValidTraceBuf2: %d\n", ValidTraceBuf2);
      }
    }

    // Get testmode indicator
    if (k_its("TestMode")) {
      testmode = atoi(k_str());

      if (testmode < 0 || testmode > 1) {
        logit("e", "Error: Invalid TestMode value.\n");
        return -1;
      }

      if (debug > 1) {
        logit("", "TestMode: %d\n", testmode);
      }
    }

    ///////////////////////////////////////////////////////////////////
    // Maybe add an "output mode" configuration parameter to specify
    // whether this is to be used in EW mode or SNW mode?
    ///////////////////////////////////////////////////////////////////

  } // end while

  // Close config file
  k_close();

  if (debug > 1) {
    logit("", "Config file closed.\n");
  }

  // Check that SCNL file has been obtained
  if (strlen(scnlFile) == 0) {
    logit("e", "Error: No SCNL input file specified.\n");
    return -1;
  }

  // Get count of SCNL tags
  fin = fopen(scnlFile, "r");

  if (debug > 1) {
    logit("", "Get SCNL tag count from %s\n", scnlFile);
  }

  if (fin == (FILE *)NULL) {
    logit("e", "Error: SCNL input file not found: %s\n", scnlFile);
    return -1;
  }

  nTag = 0;

  while (1) {
    if (fgets(line, 1024, fin) == NULL)
      break;

    ++nTag;
  }

  fclose(fin);

  // Allocate space for various arrays
  /*****************************************************************************/

  if (debug > 1) {
    logit("", "Allocate space for various arrays.\n");
  }

  // Allocate space for nSamp and sleptStatus
  nSamp = (int *)malloc(nTag * sizeof(int));
  sleptStatus = (int *)malloc(nTag * sizeof(int));

  // Allocate space for, and set, current refTime
  hrtime_ew(&d_now);

  refTime = (double *)malloc(nTag * sizeof(double));
  scnl_lastPacketTime = (double *)malloc(nTag * sizeof(double));

  for (i = 0; i < nTag; ++i) {
    refTime[i] = (int)(d_now + 1); // The next integer second value after d_now
    scnl_lastPacketTime[i] =
        (int)(d_now + 1); // The next integer second value after d_now
  }

  // Allocate space for pointers to all the inputTime and inputLat arrays
  inputTime = (double **)malloc(nTag * sizeof(double *));
  inputLat = (double **)malloc(nTag * sizeof(double *));
  feedLat = (double **)malloc(nTag * sizeof(double *));
  packet_size = (double **)malloc(nTag * sizeof(double *));

  // Allocate space for each input array and initialize counts
  for (i = 0; i < nTag; ++i) {
    nSamp[i] = 0;
    sleptStatus[i] = 0;

    inputTime[i] = (double *)malloc(1024 * sizeof(double)); // 1024 is arbitrary
    inputLat[i] = (double *)malloc(1024 * sizeof(double));  // 1024 is arbitrary
    feedLat[i] = (double *)malloc(1024 * sizeof(double));   // 1024 is arbitrary
    packet_size[i] =
        (double *)malloc(1024 * sizeof(double)); // 1024 is arbitrary
  }

  // Allocate space for tag structures
  tag = (struct TAG *)malloc(nTag * sizeof(struct TAG));

  /*****************************************************************************/
  // end array allocations

  // Parse SCNL tag
  /*****************************************************************************/

  if (debug > 1) {
    logit("", "Parse SCNL tag.\n");
  }

  fin = fopen(scnlFile, "r");

  for (i = 0; i < nTag; ++i) {
    // Initialize SCNL to empty
    strcpy(tag[i].sta, "");
    strcpy(tag[i].cha, "");
    strcpy(tag[i].net, "");
    strcpy(tag[i].loc, "");
    tag[i].mode = defaultMode;

    fgets(line, 1024, fin); // Read a line from scnlFile
    len = strlen(line);

    if (strstr(line, "RAW_MODE") != 0)
      tag[i].mode = RAW_MODE;
    else if (strstr(line, "FORCED_MODE") != 0)
      tag[i].mode = FORCED_MODE;
    else if (strstr(line, "INTERP_MODE") != 0)
      tag[i].mode = INTERP_MODE;
    else if (strstr(line, "SNW_MODE") != 0)
      tag[i].mode = SNW_MODE;

    begin = line;
    end = line;

    /********************************************************************************/
    // Procedure: Increment "end" pointer until it encounters a delimiter or
    // a break (i.e. '.' or '\n' or '\0') and explicitly set this character
    // to a null.  The substring between the "begin" and "end" pointers
    // (accessed via "begin") constitutes the value for the SCNL field
    // being read (e.g. Station).
    //
    // Check that the field value read in does not exceed the maximum allowed
    // size. Copy the value into the appropriate "tag" structure variable. Reset
    // "begin" to point to the next substring (i.e. "end + 1") and increment
    // "end" to point to the same place.
    //
    // Check to see whether "end" is  pointing beyond the end of the line being
    // read. If it is not, then repeat the process for the next substring; if it
    // is, then move on to the next line in the file.
    //
    // This procedure repeats for all four SCNL fields.
    //
    // Due to the inconsistent nature of "location" values seen in
    // input files, we place a special check that replaces values with
    // a "-" in them with an empty string.
    /********************************************************************************/

    // Get station
    while (*end != '.' && *end != '\n' && *end != '\0') {
      ++end;
    }

    *end = '\0'; // Replace delimiter with a null

    if (strlen(begin) > TRACE_STA_LEN)
      continue;

    strcpy(tag[i].sta, begin);

    begin = end + 1;
    end = begin;

    if ((end - line) >= len)
      continue;

    // Get channel
    while (*end != '.' && *end != '\n' && *end != '\0') {
      ++end;
    }

    *end = '\0'; // Replace delimiter with a null

    if (strlen(begin) > TRACE_CHAN_LEN)
      continue;

    strcpy(tag[i].cha, begin);

    begin = end + 1;
    end = begin;

    if ((end - line) >= len)
      continue;

    // Get network
    while (*end != '.' && *end != '\n' && *end != '\0') {
      ++end;
    }

    *end = '\0'; // Replace delimiter with a null

    if (strlen(begin) > TRACE_NET_LEN)
      continue;

    strcpy(tag[i].net, begin);

    begin = end + 1;
    end = begin;

    if ((end - line) >= len)
      continue;

    // Get location
    while (*end != '.' && *end != '\n' && *end != '\0') {
      ++end;
    }

    *end = '\0'; // Replace delimiter with a null

    if (strlen(begin) > TRACE_LOC_LEN)
      continue;

    strcpy(tag[i].loc, begin);

    //      if(strstr(begin, "-") != 0)
    //              strcpy(tag[i].loc, "");
  }

  /*****************************************************************************/
  // end SCNL tag parsing

  fclose(fin); // Close tag list file

  if (debug > 1) {
    for (i = 0; i < nTag; ++i) {
      logit("", "tag%d:[%s][%s][%s][%s][%d]\n", i, tag[i].sta, tag[i].cha,
            tag[i].net, tag[i].loc, tag[i].mode);
    }
  }

  // Check that Ring names have been obtained
  if (strlen(inRing) == 0) {
    logit("e", "Error: No input ring name found; exiting...\n");
    return -1;
  }
  if (strlen(outRing) == 0 && defaultMode != SNW_MODE) {
    logit("e", "Error: No ring names found; exiting...\n");
    return -1;
  }

  // Attach to input ring
  if ((inputKey = GetKey(inRing)) == -1) {
    logit("e", "Error: Invalid RingName; exiting...\n");
    return -1;
  }
  tport_attach(&region_in, inputKey);

  if (debug > 1) {
    logit("", "Attached to input ring.\n");
  }

  // Attach to output ring
  if (defaultMode != SNW_MODE) {
    if ((outputKey = GetKey(outRing)) == -1) {
      logit("e", "Error: Invalid RingName; exiting...\n");
      return -1;
    }
    tport_attach(&region_out, outputKey);

    if (debug > 1) {
      logit("", "Attached to output ring.\n");
    }
  }

  // Look up local installation id
  if (GetLocalInst(&InstId) != 0) {
    logit("e", "Error getting local installation id; exiting...\n");
    return -1;
  }

  // Specify logos to get
  if (GetType("TYPE_TRACEBUF2", &Type_TraceBuf2) != 0) {
    logit("e",
          "%s: Invalid message type <TYPE_TRACEBUF2>! Missing from earthworm.d "
          "or earthworm_global.d\n",
          argv[0]);
    return -1;
  }

  if (GetType("TYPE_HEARTBEAT", &Type_Heartbeat) != 0) {
    logit("e", "%s: Error getting Type_Heartbeat.\n", argv[0]);
    return -1;
  }

  if (GetType("TYPE_ERROR", &Type_Error) != 0) {
    logit("e", "%s: Error getting TYPE_ERROR.\n", argv[0]);
    return -1;
  }

  if (GetModId("MOD_WILDCARD", &ModWildcard) != 0) {
    return -1;
  }

  if (GetInst("INST_WILDCARD", &InstWildcard) != 0) {
    logit("e",
          "%s: Invalid instid <INST_WILDCARD>! Missing from earthworm.d or "
          "earthworm_global.d\n",
          argv[0]);
    return -1;
  }

  // Get current time
  hrtime_ew(&d_now);
  // send initial heartbeat
  status(Type_Heartbeat, 0, "");
  then = d_now;

  if (debug > 1) {
    logit("", "Construct input and output messages.\n");
  }

  // Construct input message logo
  getLogo.type = Type_TraceBuf2;
  getLogo.instid = InstWildcard;
  getLogo.mod = ModWildcard;

  // Construct output message logo
  if (defaultMode != SNW_MODE) {
    if (!typeLatencyScnl) {
      outputLogo.type = Type_TraceBuf2;
    } else {
      outputLogo.type = OutputType;
    }
    outputLogo.mod = ModId;
    outputLogo.instid = InstId;
  } else {
    sock = client_connect(
        sockName, portNum); // Change these (name and port) to config parameters

    if (sock == -1) {
      logit("", "Warning: client_connect(%s, %d) return = %d\n", sockName,
            portNum, sock);
    }

    int tbl = 1;
    if (tbl) {
      if (debug) {
        logit("", "sock = %d\n", sock);
      }

      logit("", "| NETSTA | P |           DATETIME   |       DL   |       FL   "
                "  | T_SINCE  |    P_SIZE  |\n");
    }
  }

  // Flush the ring
  while ((flag = tport_getflag(&region_in)) != TERMINATE && flag != pid &&
         tport_copyfrom(&region_in, &getLogo, nLogo, &inputLogo, &inputLength,
                        inmsg, MAX_TRACEBUF_SIZ,
                        &sequence_number) != GET_NONE) {
    // sleep_ew(10);
    // logit("", "Flushing record\n");
    // NO-OP
  }

  ///////////////////////////////////////////// OFFSET
  // TPORT_LOCATION *oldLocation;
  // TPORT_HEAD     *tmphd;
  // FILE           *foff;
  // long             offset;
  // char            offString[1024];
  ///////////////////////////////////////////// OFFSET

  // Main Loop
  while ((flag = tport_getflag(&region_in)) != TERMINATE && flag != pid) {
    // Get current time
    hrtime_ew(&d_now);

    if ((d_now - then) >= heartbeatTime) {
      // send heartbeat
      status(Type_Heartbeat, 0, "");
      then = d_now;
    }

    rc = tport_copyfrom(&region_in, &getLogo, nLogo, &inputLogo, &inputLength,
                        inmsg, MAX_TRACEBUF_SIZ, &sequence_number);
    if (rc == GET_NONE) {
      sleep_ew(sleepTime);
      if ((flag = tport_getflag(&region_in)) == TERMINATE || flag == pid) {
        break;
      }
      slept = 1;

      for (i = 0; i < nTag; ++i) {
        sleptStatus[i] = 1;
      }

      if (debug > 1) {
        logit("t", "Slept for %f.\n", sleepTime);
      }

      // After the sleep we want to double check to see if something should be
      // output in the Forced and Interpolated modes before we overwrite it with
      // a new record

      if (defaultMode == SNW_MODE) {
        getSNWLatency();
      } else {
        getForcedLatency(); // Output latency values forced to 1-sec intervals
        getInterpolatedLatency(); // Calculate latency values for archiving
      }

      continue;
    }

    if (rc == GET_TOOBIG) {
      logit("t", "Warning: retrieved message too big (%ld) for inmsg\n",
            inputLength);
      continue;
    }

    if (rc == GET_NOTRACK)
      logit("t", "Warning: Tracking error.\n");

    if (rc == GET_MISS_LAPPED)
      logit("t", "Warning: Got lapped on the ring.\n");

    if (rc == GET_MISS_SEQGAP)
      logit("t", "Warning: Gap in sequence numbers\n");

    if (rc == GET_MISS)
      logit("t", "Warning: Missed messages\n");

    strcpy(orig_datatype, trh->datatype);
    if (WaveMsg2MakeLocal(trh) < 0) {
      char scnl[20], dt[3];
      scnl[0] = 0;
      strcat(scnl, trh->sta);
      strcat(scnl, ".");
      strcat(scnl, trh->chan);
      strcat(scnl, ".");
      strcat(scnl, trh->net);
      strcat(scnl, ".");
      strcat(scnl, trh->loc);
      /* now put a space if there are any punctuation marks */

      for (i = 0; i < 15; i++) {
        if (!isalnum(scnl[i]) && !ispunct(scnl[i]))
          scnl[i] = ' ';
      }
      strncpy(dt, trh->datatype, 2);
      for (i = 0; i < 2; i++) {
        if (!isalnum(dt[i]) && !ispunct(dt[i]))
          dt[i] = ' ';
      }
      dt[i] = 0;
      logit("t",
            "WARNING: WaveMsg2MakeLocal rejected tracebuf.  Discard (%s).\n",
            scnl);
      logit("", "\tdatatype=[%s]\n", dt);
      continue;
    }

    // Print input header
    if (debug > 2) {
      logit("t", "Input Record --> ");
      logit("", "%5s.%s.%s.%s (0x%x 0x%x) ", trh->sta, trh->chan, trh->net,
            trh->loc, trh->version[0], trh->version[1]);
      logit("", "%d %s %3d ", trh->pinno, orig_datatype, trh->nsamp);
      logit("", "%6.4f", trh->samprate);
      logit("", " (%.4f) (%.4f)\n", trh->starttime, trh->endtime);
    }

    // Find the inputTime and inputLat arrays that match current record tag

    iTag = -1;

    for (i = 0; i < nTag; ++i) {
      if (strcmp(trh->sta, tag[i].sta) == 0 &&
          strcmp(trh->chan, tag[i].cha) == 0 &&
          strcmp(trh->net, tag[i].net) == 0 &&
          ((strcmp(trh->loc, tag[i].loc) == 0) ||
           (strcmp(tag[i].loc, "") == 0))) {
        iTag = i; // If match found for SCNL tag, set value and break
        break;
      }
    }

    if (iTag == -1)
      continue; // If no input is found for this tag at this time,
                // continue to next loop iteration

    //      Get current time
    hrtime_ew(&d_now);

    // Testing and debugging clause: set t_offset
    if (testmode && t_offset == 0) {
      t_offset = d_now - trh->endtime;

      if (debug > 1) {
        logit("t", "t_offset: %f\n", t_offset);
      }
    }

    // Compute latencies
    DL = d_now - (trh->endtime + t_offset);
    TL = d_now - (trh->starttime + t_offset);
    ML = (DL + TL) / 2.;
    FL = d_now - scnl_lastPacketTime[iTag];

    // Add read values to our arrays
    // >>> arrayName[index of current tag][current count associated with that
    // tag]
    inputTime[iTag][nSamp[iTag]] = trh->endtime;
    inputLat[iTag][nSamp[iTag]] = DL;
    scnl_lastPacketTime[iTag] = d_now;
    packet_size[iTag][nSamp[iTag]] = (trh->endtime - trh->starttime);
    feedLat[iTag][nSamp[iTag]] = FL;

    if (debug > 1) {
      logit("t", "\nMain loop: iTag=%d, nSamp[iTag]=%d   -->  ", iTag,
            nSamp[iTag]);
      logit("", "DL=%f, FL=%f\n", DL, FL);
    }

    ++nSamp[iTag]; // Increment count associated with that particular tag only

    if (tag[i].mode == RAW_MODE)
      getRawLatency(); // Output real-time latency values

    // Process these regardless (to see whether any of the tags
    // marked with these modes are in the middle of a gap)

    if (defaultMode == SNW_MODE) {
      getSNWLatency();
    } else {
      getForcedLatency(); // Output latency values forced to 1-sec intervals
      getInterpolatedLatency(); // Calculate latency values for archiving
    }
  } // end while

  return 0;
} // end main

/************************************************************************************/

/******************************************************************************
 * status() builds a heartbeat or error message & puts it into        *
 *                   shared memory.  Writes errors to log file & screen.      *
 ******************************************************************************/
void status(unsigned char type, short ierr, char *note) {
  MSG_LOGO logo;
  char msg[256];
  long size;
  time_t t;

  /* Build the message
   *******************/
  logo.instid = InstId;
  logo.mod = ModId;
  logo.type = type;

  time(&t);

  if (type == Type_Heartbeat) {
    sprintf(msg, "%ld %ld\n", (long)t, (long)pid);
    if (debug)
      logit("t", "status: %s: %s\n", ProgName, msg);
  } else if (type == Type_Error) {
    sprintf(msg, "%ld %hd %s\n", (long)t, ierr, note);
    logit("et", "status: %s: %s\n", ProgName, note);
  }

  size = strlen(msg); /* don't include the null byte in the message */

  /* Write the message to shared memory
   ************************************/
  if (tport_putmsg(&region_in, &logo, size, msg) != PUT_OK) {
    if (type == Type_Heartbeat) {
      logit("et", "%s:  Error sending heartbeat.\n", ProgName);
    } else if (type == Type_Error) {
      logit("et", "%s:  Error sending error:%d.\n", ProgName, ierr);
    }
  }
  return;
}

/************************************************************************************/

/************************************************************************************/
/* Function: writeOutputMessage */
/************************************************************************************/

static void writeOutputMessage(void) {
  if (typeLatencyScnl == 1 || typeLatencyScnl == 2) {
    int fixed;
    if (typeLatencyScnl == 1)
      fixed = 1;
    else
      fixed = 0;
    char latency_scnl_buf[LATENCY_STR_SIZE];
    tbuf2latencystr(trout, *long_data_out, latency_scnl_buf, fixed);
    tport_putmsg(&region_out, &outputLogo, strlen(latency_scnl_buf),
                 latency_scnl_buf);
    if (debug) {
      logit("t", "Output Latency --> %s\n", latency_scnl_buf);
    }
  } else if (typeLatencyScnl == 3) {
    LATENCY_SCNL_BIN latency_scnl_bin; // latency scnl binary
    tbuf2latencybin(trout, *long_data_out, &latency_scnl_bin);
    tport_putmsg(&region_out, &outputLogo, sizeof(LATENCY_SCNL_BIN),
                 (char *)&latency_scnl_bin);
  } else {
    if (ValidTraceBuf2) {
      trout->endtime = trout->starttime;
    }
    tport_putmsg(&region_out, &outputLogo, outputLength, outmsg);
  }
}

/************************************************************************************/

/************************************************************************************/
/* Function: getRawLatency */
/************************************************************************************/

int getRawLatency() {
  int j; // Utility counter

  if (debug > 2) {
    logit("t", "getRawLatency()\n");
  }

  for (j = 0; j < nTag; ++j) // 'j' is the index of the tag we are considering
  {
    if (nSamp[j] > 0 && tag[j].mode == RAW_MODE) {

      // Convert output latency to integer rounded to the nearest millisecond
      *long_data_out = (int)(1000 * inputLat[j][0] + 0.5);

      // Populate trout and apply naming convention rules
      strcpy(trout->sta, tag[j].sta);
      strcpy(trout->chan, tag[j].cha);
      strcpy(trout->net, tag[j].net);
      strcpy(trout->loc, tag[j].loc);
      if (LocationCode[0] != 0) {
        strcpy(trout->loc, LocationCode);
      }
      if (renameChannel) {
        trout->chan[0] = chan0;
        if (trout->chan[1] == 'H') {
          trout->chan[1] = chan1;
        }
        if (trout->chan[1] == 'N') {
          trout->chan[1] = chan1;
        }
      }

      trout->version[0] = v0;
      trout->version[1] = v1;
      trout->pinno = pin;
      trout->nsamp = outNsamp;
      trout->samprate = samprate;
      // t_offset makes it relative to d_now; delete later
      trout->endtime = inputTime[j][0] + t_offset;
      if (typeLatencyScnl) {
        trout->starttime = trout->endtime - packet_size[j][0];
      } else {
        trout->starttime = trout->endtime - increment;
      }
      strcpy(trout->datatype, datatype);
      // strcpy(trout->quality, quality);
      strcpy(trout->pad, pad);

      outputLength =
          sizeof(TRACE2_HEADER) +
          4 * trout->nsamp; // floating-point numbers  are 4 bytes eac?

      // Print output header
      if (debug) {
        logit("t", "Raw Latency --> ");
        logit("", "%5s.%s.%s.%s (0x%x 0x%x) ", trout->sta, trout->chan,
              trout->net, trout->loc, trout->version[0], trout->version[1]);
        logit("", "%d %s %3d ", trout->pinno, trout->datatype, trout->nsamp);
        logit("", "%6.4f", trout->samprate);
        logit("", " (%.4f) (%.4f) ", trout->starttime, trout->endtime);
        logit("", "%d", long_data_out[0]);
        logit("", " [DL:%fs, FL:%fs] ", inputLat[j][0], feedLat[j][0]);

        if (debug > 1) {
          logit("", " --> sleptStatus[%d]=%d ", j, sleptStatus[j]);
        }

        if (slept == 1) {
          logit("", " Slept for %f ", sleepTime);
        }

        logit("", "\n");
      }

      if (slept == 1)
        slept = 0;

      trout->quality[0] = 0;
      trout->quality[1] = 0;

      if (sleptStatus[j] == 1)
        trout->quality[1] = 1;

      // Write output message to latency ring
      writeOutputMessage();

      sleptStatus[j] = 0;
      nSamp[j] = 0;

    } // end if
  }   // end for (each tag)
  return 0;
} // end getRawLatency()

/************************************************************************************/

/************************************************************************************/
/* Function: getForcedLatency */
/************************************************************************************/

int getForcedLatency() {
  int j; // Utility counter

  if (debug > 2) {
    logit("t", "getForcedLatency(), ");
    logit("", "  d_now: %f\n", d_now);
  }

  hrtime_ew(&d_now);

  for (j = 0; j < nTag; ++j) // 'j' is the index of the tag we are considering
  {
    if (nSamp[j] == 0) {
      continue;
    }

    if (d_now - (inputTime[j][nSamp[j] - 1] + t_offset) > timeout) {

      if (debug > 2) {
        logit("", "  d_now: %f\n", d_now);
        logit("", "  t_offset: %f\n", t_offset);
        logit("", "  inputTime[j][nSamp[j]-1]: %f\n",
              inputTime[j][nSamp[j] - 1]);
      }

      nSamp[j] = 0;

      if (debug > 1) {
        logit("t", "Timeout reached; continuing...\n");
      }

      continue;
    }

    if (nSamp[j] > 0 && tag[j].mode == FORCED_MODE) {
      if (refTime[j] > d_now) {
        if (debug > 1) {
          logit("t", "refTime[%d]=%f, d_now=%f\n", j, refTime[j], d_now);
        }

        break;
      }

      // Convert output latency to integer rounded to the nearest millisecond
      *long_data_out = (int)(1000 * inputLat[j][nSamp[j] - 1] + 0.5);

      // Populate trout and apply naming convention rules
      strcpy(trout->sta, tag[j].sta);
      strcpy(trout->chan, tag[j].cha);
      strcpy(trout->net, tag[j].net);
      strcpy(trout->loc, tag[j].loc);
      if (LocationCode[0] != 0) {
        strcpy(trout->loc, LocationCode);
      }

      if (renameChannel) {
        trout->chan[0] = chan0;
        if (trout->chan[1] == 'H') {
          trout->chan[1] = chan1;
        }
        if (trout->chan[1] == 'N') {
          trout->chan[1] = chan1;
        }
      }

      trout->version[0] = v0;
      trout->version[1] = v1;
      trout->pinno = pin;
      trout->nsamp = outNsamp;
      trout->samprate = samprate;
      trout->endtime = refTime[j];
      trout->starttime = trout->endtime - increment;
      strcpy(trout->datatype, datatype);
      // strcpy(trout->quality, quality);
      strcpy(trout->pad, pad);

      outputLength = sizeof(TRACE2_HEADER) + 4 * trout->nsamp;

      // Print output header
      if (debug) {
        logit("t", "Forced Latency --> ");

        logit("", "%5s.%s.%s.%s (0x%x 0x%x) ", trout->sta, trout->chan,
              trout->net, trout->loc, trout->version[0], trout->version[1]);
        logit("", "%d %s %3d ", trout->pinno, trout->datatype, trout->nsamp);
        logit("", "%6.4f", trout->samprate);
        logit("", " (%.4f) (%.4f) ", trout->starttime, trout->endtime);
        logit("", "%8d", long_data_out[0]);
        logit("", " [DL:%fs, FL:%fs] ", inputLat[j][nSamp[j] - 1],
              feedLat[j][nSamp[j] - 1]);

        if (slept == 1) {
          logit("", " Slept for %f ", sleepTime);
        }

        logit("", "\n");
      }

      if (slept == 1)
        slept = 0;

      trout->quality[0] = 0;
      trout->quality[1] = 0;

      if (sleptStatus[j] == 1)
        trout->quality[1] = 1;

      //      if(sleptStatus[j] == 1)
      //              strcpy(trout->quality, "1");

      // Write output message to latency ring
      writeOutputMessage();

      sleptStatus[j] = 0;

      inputTime[j][0] = inputTime[j][nSamp[j] - 1];
      inputLat[j][0] = inputLat[j][nSamp[j] - 1];

      nSamp[j] = 1;

      // Increment the reference time
      refTime[j] += increment;

      // If we do not have data to fill in intermediate times,
      // shift the reference time that we are looking for.
      while (refTime[j] < (inputTime[j][0] + t_offset)) {
        trout->endtime = refTime[j];
        trout->starttime = trout->endtime - increment;

        // Write output message to latency ring
        writeOutputMessage();

        refTime[j] += increment;
      }

      if (debug > 1) {
        logit("", "refTime[%d]=%f\n", j, refTime[j]);
      }

    } // end if
  }   // end for (each tag)
  return 0;
} // end getForcedLatency()

/************************************************************************************/

/************************************************************************************/
/* Function: getInterpolatedLatency */
/************************************************************************************/

int getInterpolatedLatency() {
  int nShift;          // Count used to discard already-processed inputs
  int iBefore, iAfter; // Indices of inputs bracketing interval of interest
  int nOut;            // Utility counter
  int i, j;            // Utility counter
  double tStart, tEnd; // Start/end times for the interval being integrated over
  double time[1024];   // Times for point being integrated over
  double latency[1024]; // Latencies for point being integrated over
  double outputLatency; // Latency value calculated for a given refTime
  double frac;          // Fractional value used interpolation calculation

  if (debug > 2) {
    logit("t", "getInterpolatedLatency()\n");
  }

  hrtime_ew(&d_now);

  for (j = 0; j < nTag; ++j) // 'j' is the index of the tag we are considering
  {
    if (tag[j].mode != INTERP_MODE)
      continue;

    if (nSamp[j] == 0)
      continue;

    if (d_now - (inputTime[j][nSamp[j] - 1] + t_offset) > timeout) {
      nSamp[j] = 0;
      continue;
    }

    while (1) {
      // Set interval start and end for the iteration based on refTime
      tStart = refTime[j] - 0.5 * interval;
      tEnd = refTime[j] + 0.5 * interval;

      if (tStart < (inputTime[j][0] + t_offset)) {
        refTime[j] += interval;

        if (debug > 1) {
          logit("t", "First point, invalid tStart; continuing...\n");
        }

        continue;
      }

      if (debug > 1) {
        logit("t", "refTime[%d]: %11.2f\n", j, refTime[j]);
        logit("", "tStart: %11.2f\n", tStart);
        logit("", "tEnd:   %11.2f\n", tEnd);
      }

      // Analyze arrays to determine whether we can output anything

      iBefore = 0;
      iAfter = 0;

      for (i = 0; i < nSamp[j];
           ++i) // For the records currently in our sample window
      {
        if ((inputTime[j][i] + t_offset) <= tStart) {
          // gotBefore = 1;

          iBefore = i; // Index of the last data value before (or at) tStart

          if (debug > 2)
            logit("", "inputTime[%d][%d] + t_offset <= tStart; iBefore=%d\n", j,
                  i, i);
        }
      }

      // Shift values in the arrays (discard values that are not necessary
      // for future processing, to save space).  Any future values of 'tStart'
      // will be later than the current value of 'tStart', which is itself later
      // than the time pointed to by 'iBefore'.  Records preceding 'iBefore'
      // are therefore unnecessary.

      nShift = iBefore;

      for (i = nShift; i < nSamp[j]; ++i) {
        inputTime[j][i - nShift] = inputTime[j][i];
        inputLat[j][i - nShift] = inputLat[j][i];
      }

      nSamp[j] -= nShift; // Shift sampling window
      iBefore = 0;        // Reset index once shift has been performed

      for (i = 0; i < nSamp[j];
           ++i) // For the records currently in our sample window
      {
        if ((inputTime[j][i] + t_offset) >= tEnd) {
          iAfter = i; // Index of the first data value after (or at) tEnd

          if (debug > 2)
            logit("",
                  "inputTime[%d][%d] + t_offset >= tEnd; iAfter=%d; breaking\n",
                  j, i, i);

          break;
        }
      }

      if (iBefore == iAfter && iAfter < (nSamp[j] - 1))
        ++iAfter;

      if (debug > 1) {
        logit("", "Held Input Times: ");

        for (i = 0; i < nSamp[j]; ++i)
          logit("", "%11.2f ", (inputTime[j][i] + t_offset));

        logit("", "\n");

        logit("", "iBefore=%d, iAfter=%d\n", iBefore, iAfter);
      }

      if ((inputTime[j][iAfter] + t_offset) <
          tEnd) // If pointer for iAfter is not beyond tEnd,
        break;  // then we do not have enough data to do the
      // next interpolation, so we need to break
      // out of this function and wait for another
      // read.

      // Calculate interpolation weighting fraction for beginning of time window
      frac = (tStart - (inputTime[j][iBefore] + t_offset)) /
             (inputTime[j][iBefore + 1] - inputTime[j][iBefore]);

      if (debug > 1) {
        logit("", "frac(begin)=%f\n", frac);
      }

      if (frac < 0 || frac > 1.0) {
        logit("e", "Error: Invalid fraction ( tStart ).\n");
        exit(0);
      }

      nOut = 0; // Reset counter at beginning of time window

      // Apply weighting and add interpolated latency to array used for
      // calculations
      time[nOut] = tStart;
      latency[nOut] =
          frac * inputLat[j][iBefore + 1] + (1. - frac) * inputLat[j][iBefore];
      ++nOut;

      // Add points entirely within time window to the arrays used for
      // calculations
      for (i = iBefore + 1; i < iAfter; ++i) {
        time[nOut] = inputTime[j][i] + t_offset;
        latency[nOut] = inputLat[j][i];
        ++nOut;
      }

      // Re-calculate interpolation weighting fraction for end of time window
      frac = (tEnd - (inputTime[j][iAfter - 1] + t_offset)) /
             (inputTime[j][iAfter] - inputTime[j][iAfter - 1]);

      if (debug > 1) {
        logit("", "frac(end)=%f\n", frac);
      }

      if (frac < 0 || frac > 1.0) {
        logit("e", "Error: Invalid fraction ( tEnd ).\n");
        exit(0);
      }

      // Apply weighting and add interpolated latency to array used for
      // calculations
      time[nOut] = tEnd;
      latency[nOut] =
          frac * inputLat[j][iAfter] + (1. - frac) * inputLat[j][iAfter - 1];
      ++nOut;

      if (debug > 1) {
        for (i = 0; i < nOut; ++i) {
          logit("", "time[%d]=%f, latency[%d]=%f\n", i, time[i], i, latency[i]);
        }
      }

      // Calculate mean latency as trapezoidal summation over current time
      // window
      outputLatency = 0;

      for (i = 1; i < nOut; ++i) {
        outputLatency +=
            0.5 * (latency[i] + latency[i - 1]) * (time[i] - time[i - 1]);
      }

      outputLatency /= interval;

      if (debug > 1) {
        logit("",
              "Tag=%d, iBefore=%d, iAfter=%d: refTime=%11.2f, outputLatency= "
              "%7.4f, nshift=%d\n",
              j, iBefore, iAfter, refTime[j], outputLatency, iBefore + 1);
      }

      if (debug > 1) {
        logit("", "Time: %11.2f, Latency: %7.4f\n", refTime[j], outputLatency);
      }

      // Convert output latency to integer rounded to the nearest millisecond
      *long_data_out = (int)(1000 * outputLatency + 0.5);

      // Populate trout and apply naming convention rules
      strcpy(trout->sta, tag[j].sta);
      strcpy(trout->chan, tag[j].cha);
      strcpy(trout->net, tag[j].net);
      strcpy(trout->loc, tag[j].loc);
      if (LocationCode[0] != 0) {
        strcpy(trout->loc, LocationCode);
      }

      if (renameChannel) {
        trout->chan[0] = chan0;
        if (trout->chan[1] == 'H') {
          trout->chan[1] = chan1;
        }
        if (trout->chan[1] == 'N') {
          trout->chan[1] = chan1;
        }
      }

      trout->version[0] = v0;
      trout->version[1] = v1;
      trout->pinno = pin;
      trout->nsamp = outNsamp;
      trout->samprate = samprate;
      trout->endtime = refTime[j];
      trout->starttime = trout->endtime - increment;
      strcpy(trout->datatype, datatype);
      // strcpy(trout->quality, quality);
      strcpy(trout->pad, pad);

      outputLength = sizeof(TRACE2_HEADER) + 4 * trout->nsamp;

      // Print output header
      if (debug) {
        logit("", "Interpolated Latency --> ");
        logit("", "%5s.%s.%s.%s (0x%x 0x%x) ", trout->sta, trout->chan,
              trout->net, trout->loc, trout->version[0], trout->version[1]);
        logit("", "%d %s %3d ", trout->pinno, trout->datatype, trout->nsamp);
        logit("", "%6.4f", trout->samprate);
        logit("", " (%.4f) (%.4f) ", trout->starttime, trout->endtime);
        logit("", "%d", long_data_out[0]);

        if (slept == 1) {
          logit("", " Slept for %f ", sleepTime);
        }

        logit("", "\n");
      }

      if (slept == 1)
        slept = 0;

      trout->quality[0] = 0;
      trout->quality[1] = 0;

      if (sleptStatus[j] == 1)
        trout->quality[1] = 1;

      //      if(sleptStatus[j] == 1)
      //              strcpy(trout->quality, "1");

      // Write output message to latency ring
      writeOutputMessage();

      sleptStatus[j] = 0;

      // Increment the reference time
      refTime[j] += increment;

      // If we do not have data to fill in intermediate times,
      // shift the reference time that we are looking for.
      if (refTime[j] < (inputTime[j][0] + t_offset)) {
        refTime[j] = (int)((inputTime[j][0] + t_offset) + increment);
      }
    } // end while (processing a time window)
  }   // end for (each tag)
  return 0;
} // end getInterpolatedLatency

/************************************************************************************/

/************************************************************************************/
/* Function: getSNWLatency */
/************************************************************************************/

int getSNWLatency() {
  int nShift;          // Count used to discard already-processed inputs
  int iBefore, iAfter; // Indices of inputs bracketing interval of interest
  int nOut;            // Utility counter
  int i, j;            // Utility counter
  double tStart, tEnd; // Start/end times for the interval being integrated over
  double time[1024];   // Times for point being integrated over
  double dlatency[1024]; // Data latencies (DL) for point being integrated over
  double flatency[1024]; // Feed latencies (FL) for point being integrated over
  double psize[1024];    // Packet sizes for point being integrated over
  double outputDLatency; // Data latency value calculated for a given refTime
  double outputFLatency; // Feed latency value calculated for a given refTime
  double avPacketSize;   // Average packet size
  double frac;           // Fractional value used interpolation calculation
  double time_since_last = 0.0;
  double flatmin = 0.01;

  if (debug > 2) {
    logit("t", "getSNWLatency()\n");
  }

  hrtime_ew(&d_now);

  for (j = 0; j < nTag; ++j) // 'j' is the index of the tag we are considering
  {
    if (tag[j].mode != SNW_MODE)
      continue;

    if (nSamp[j] == 0)
      continue;

    if (d_now - (inputTime[j][nSamp[j] - 1] + t_offset) > timeout) {
      nSamp[j] = 0;
      continue;
    }

    while (1) {
      // Set interval start and end for the iteration based on refTime
      tStart = refTime[j] - 0.5 * interval;
      tEnd = refTime[j] + 0.5 * interval;

      if (tStart < (inputTime[j][0] + t_offset)) {
        refTime[j] += interval;

        if (debug > 1) {
          logit("t", "First point, invalid tStart; continuing...\n");
        }

        continue;
      }

      if (debug > 1) {
        logit("t", "refTime[%d]: %11.2f\n", j, refTime[j]);
        logit("", "tStart: %11.2f\n", tStart);
        logit("", "tEnd:   %11.2f\n", tEnd);
      }

      // Analyze arrays to determine whether we can output anything

      iBefore = 0;
      iAfter = 0;

      for (i = 0; i < nSamp[j];
           ++i) // For the records currently in our sample window
      {
        if ((inputTime[j][i] + t_offset) <= tStart) {
          // gotBefore = 1;

          iBefore = i; // Index of the last data value before (or at) tStart

          if (debug > 2)
            logit("", "inputTime[%d][%d] + t_offset <= tStart; iBefore=%d\n", j,
                  i, i);
        }
      }

      // Shift values in the arrays (discard values that are not necessary
      // for future processing, to save space).  Any future values of 'tStart'
      // will be later than the current value of 'tStart', which is itself later
      // than the time pointed to by 'iBefore'.  Records preceding 'iBefore'
      // are therefore unnecessary.

      nShift = iBefore;

      for (i = nShift; i < nSamp[j]; ++i) {
        inputTime[j][i - nShift] = inputTime[j][i];
        inputLat[j][i - nShift] = inputLat[j][i];
        feedLat[j][i - nShift] = feedLat[j][i];
      }

      nSamp[j] -= nShift; // Shift sampling window
      iBefore = 0;        // Reset index once shift has been performed

      for (i = 0; i < nSamp[j];
           ++i) // For the records currently in our sample window
      {
        if ((inputTime[j][i] + t_offset) >= tEnd) {
          iAfter = i; // Index of the first data value after (or at) tEnd

          if (debug > 2)
            logit("",
                  "inputTime[%d][%d] + t_offset >= tEnd; iAfter=%d; breaking\n",
                  j, i, i);

          break;
        }
      }

      if (iBefore == iAfter && iAfter < (nSamp[j] - 1))
        ++iAfter;

      if (debug > 1) {
        logit("", "Held Input Times: ");

        for (i = 0; i < nSamp[j]; ++i)
          logit("", "%11.2f ", (inputTime[j][i] + t_offset));

        logit("", "\n");

        logit("", "iBefore=%d, iAfter=%d\n", iBefore, iAfter);
      }

      if ((inputTime[j][iAfter] + t_offset) <
          tEnd) // If pointer for iAfter is not beyond tEnd,
        break;  // then we do not have enough data to do the
      // next interpolation, so we need to break
      // out of this function and wait for another
      // read.

      // Calculate interpolation weighting fraction for beginning of time window
      frac = (tStart - (inputTime[j][iBefore] + t_offset)) /
             (inputTime[j][iBefore + 1] - inputTime[j][iBefore]);

      if (debug > 1) {
        logit("", "frac(begin)=%f\n", frac);
      }

      if (frac < 0 || frac > 1.0) {
        logit("e", "Error: Invalid fraction ( tStart ).\n");
        exit(0);
      }

      nOut = 0; // Reset counter at beginning of time window

      // Integration/averaging of feed latencies requires specialized
      // calculation due to the bizarre practice of splitting some of the larger
      // incoming packets into multiple smaller sub-packets prior to
      // transmitting them to the wave ring but only attaching the feed latency
      // of the original packet to the first of the sub-packets (which means
      // that the other sub-packets are listed as having a feed latency of 0.0,
      // which is not reflective of reality).

      // Apply weighting and add interpolated values to arrays used for
      // calculations
      time[nOut] = tStart;
      dlatency[nOut] =
          frac * inputLat[j][iBefore + 1] + (1. - frac) * inputLat[j][iBefore];
      psize[nOut] = frac * packet_size[j][iBefore + 1] +
                    (1. - frac) * packet_size[j][iBefore];

      if (feedLat[j][iBefore + 1] < flatmin && feedLat[j][iBefore] >= flatmin) {
        flatency[nOut] = feedLat[j][iBefore];
      } else if (feedLat[j][iBefore + 1] >= flatmin &&
                 feedLat[j][iBefore] < flatmin) {
        flatency[nOut] = feedLat[j][iBefore + 1];
      } else {
        flatency[nOut] =
            frac * feedLat[j][iBefore + 1] + (1. - frac) * feedLat[j][iBefore];
      }

      ++nOut;

      // Add points entirely within time window to the arrays used for
      // calculations
      for (i = iBefore + 1; i < iAfter; ++i) {
        time[nOut] = inputTime[j][i] + t_offset;
        dlatency[nOut] = inputLat[j][i];

        if (debug > 2) {
          logit("", "feedLat[%d][%d] = %f\n", j, i, feedLat[j][i]);
        }

        flatency[nOut] = feedLat[j][i];
        psize[nOut] = packet_size[j][i];
        ++nOut;
      }

      // Re-calculate interpolation weighting fraction for end of time window
      frac = (tEnd - (inputTime[j][iAfter - 1] + t_offset)) /
             (inputTime[j][iAfter] - inputTime[j][iAfter - 1]);

      //////////////////////////////////////////////////
      time_since_last = tEnd - inputTime[j][iAfter - 1];
      //////////////////////////////////////////////////

      if (debug > 1) {
        logit("", "frac(end)=%f\n", frac);
      }

      if (frac < 0 || frac > 1.0) {
        logit("e", "Error: Invalid fraction ( tEnd ).\n");
        exit(0);
      }

      // Apply weighting and add interpolated values to arrays used for
      // calculations
      time[nOut] = tEnd;
      dlatency[nOut] =
          frac * inputLat[j][iAfter] + (1. - frac) * inputLat[j][iAfter - 1];

      if (feedLat[j][iAfter + 1] < flatmin && feedLat[j][iAfter] >= flatmin)
        flatency[nOut] = feedLat[j][iAfter];
      else if (feedLat[j][iAfter + 1] >= flatmin &&
               feedLat[j][iAfter] < flatmin)
        flatency[nOut] = feedLat[j][iAfter + 1];
      else
        flatency[nOut] =
            frac * feedLat[j][iAfter] + (1. - frac) * feedLat[j][iAfter - 1];

      psize[nOut] = frac * packet_size[j][iAfter] +
                    (1. - frac) * packet_size[j][iAfter - 1];
      ++nOut;

      if (debug > 1) {
        for (i = 0; i < nOut; ++i) {
          logit("", "time[%d]=%f, dlatency[%d]=%f, psize[%d]=%f \n", i, time[i],
                i, dlatency[i], i, psize[i]);
          // logit("","time[%d]=%f, dlatency[%d]=%f, flatency[%d]=%f,
          // psize[%d]=%f \n",
          //              i, time[i], i, dlatency[i], i, flatency[i], i,
          //              psize[i]);
        }
      }

      // Calculate mean latency as trapezoidal summation over current time
      // window
      outputDLatency = 0;
      outputFLatency = 0;
      avPacketSize = 0;

      for (i = 1; i < nOut; ++i) {
        outputDLatency +=
            0.5 * (dlatency[i] + dlatency[i - 1]) * (time[i] - time[i - 1]);
        avPacketSize +=
            0.5 * (psize[i] + psize[i - 1]) * (time[i] - time[i - 1]);
      }

      outputDLatency /= interval;
      avPacketSize /= interval;

      /////////////////////////////////////////////////////////////////////////////////////////////////////////

      // The first nonzero flatency is fref.  Find on first pass.
      // On second pass, if flatency is 0, check to see if previous
      // flatency value is nonzero, and set to that.  Otherwise,
      // set to fref.  Then proceed with the summation as normal.

      double fref = 0.0;
      int frefset = 0;

      for (i = 0; i < nOut; ++i) {
        if (!frefset) {
          if (flatency[i] > flatmin) {
            fref = flatency[i];
            // logit("","fref set: %s-%s fref=%f\n", tag[j].net, tag[j].sta,
            // fref);
            frefset = 1;
          }
        }
      }

      if (flatency[0] < flatmin)
        flatency[0] = fref;

      for (i = 1; i < nOut; ++i) {
        if (flatency[i] < flatmin) {
          if (flatency[i - 1] >= flatmin) {
            flatency[i] = flatency[i - 1];
            // logit("","flatency[%d]=flatency[%d]=%f\n", i, i-1, flatency[i]);
          } else {
            flatency[i] = fref;
            // logit("","flatency[%d]=%f (fref=%f)\n", i, flatency[i], fref);
          }
        }
      }

      // logit("","%s-%s fref=%f\n", tag[j].net, tag[j].sta, fref);

      for (i = 1; i < nOut; ++i) {
        outputFLatency +=
            0.5 * (flatency[i] + flatency[i - 1]) * (time[i] - time[i - 1]);
      }
      outputFLatency /= interval;

      /////////////////////////////////////////////////////////////////////////////////////////////////////////

      if (debug > 1) {
        logit("",
              "Tag=%d, iBefore=%d, iAfter=%d: refTime=%11.2f, outputDLatency= "
              "%7.4f, outputFLatency=%7.4f, nshift=%d\n",
              j, iBefore, iAfter, refTime[j], outputDLatency, outputFLatency,
              iBefore + 1);
      }

      if (debug) {
        logit("", "Time: %11.2f, DLatency: %7.4f, FLatency: %7.4f\n",
              refTime[j], outputDLatency, outputFLatency);
      }

      // Convert output latency to integer rounded to the nearest millisecond
      *long_data_out = (int)(1000 * outputDLatency + 0.5);

      // Populate trout and apply naming convention rules
      strcpy(trout->sta, tag[j].sta);
      strcpy(trout->chan, tag[j].cha);
      strcpy(trout->net, tag[j].net);
      strcpy(trout->loc, tag[j].loc);
      if (LocationCode[0] != 0) {
        strcpy(trout->loc, LocationCode);
      }

      if (renameChannel) {
        trout->chan[0] = chan0;
        if (trout->chan[1] == 'H') {
          trout->chan[1] = chan1;
        }
        if (trout->chan[1] == 'N') {
          trout->chan[1] = chan1;
        }
      }

      trout->version[0] = v0;
      trout->version[1] = v1;
      trout->pinno = pin;
      trout->nsamp = outNsamp;
      trout->samprate = samprate;
      trout->endtime = refTime[j];
      trout->starttime = trout->endtime - increment;
      strcpy(trout->datatype, datatype);
      // strcpy(trout->quality, quality);
      strcpy(trout->pad, pad);

      outputLength = sizeof(TRACE2_HEADER) + 4 * trout->nsamp;

      // Print output header
      if (debug) {
        logit("", "SNW Latency --> ");
        logit("", "%5s.%s.%s.%s (0x%x 0x%x) ", trout->sta, trout->chan,
              trout->net, trout->loc, trout->version[0], trout->version[1]);
        logit("", "%d %s %3d ", trout->pinno, trout->datatype, trout->nsamp);
        logit("", "%6.4f", trout->samprate);
        logit("", " (%.4f) (%.4f) ", trout->starttime, trout->endtime);
        // logit("", "%d", long_data_out[0]);
        logit("", " Host:[%s] ", hostName);
        logit("", " Packet Size:[%.1f] ", avPacketSize);
        // logit("", " Time Since Last:[] ");
        logit("", " Average Latencies:[DL:%.1fs, FL %.1fs] ", outputDLatency,
              outputFLatency);

        if (slept == 1) {
          logit("", " Slept for %f ", sleepTime);
        }

        logit("", "\n");
      }

      if (slept == 1)
        slept = 0;

      trout->quality[0] = 0;
      trout->quality[1] = 0;

      if (sleptStatus[j] == 1)
        trout->quality[1] = 1;

      // logit("","d_now=%f ", d_now);
      // logit("","scnl_lastPacketTime[%d]=%f\n", j, scnl_lastPacketTime[j]);
      // logit("","\noutputFLatency=%f\n", outputFLatency);

      int parts = 0;
      char outStr[4096];
      char tempStr[1024];
      strcpy(outStr, "");

      if (avPacketSize && ((avPacketSize * 100.0) >= 5.0))
        ++parts;
      if (outputDLatency && ((outputDLatency * 100.0) >= 5.0))
        ++parts;
      if (time_since_last && time_since_last > 0.0)
        ++parts;
      if (outputFLatency && ((outputFLatency * 100.0) >= 5.0))
        ++parts;

      logit("", "SNW STRING ==>  ");
      sprintf(tempStr, "%s-%s:%d:", trout->net, trout->sta, parts);
      strcat(outStr, tempStr);

      if (avPacketSize && ((avPacketSize * 100.0) >= 5.0)) {
        sprintf(tempStr, "\"%s - Packet Size\"=%.3f;", hostName, avPacketSize);
        strcat(outStr, tempStr);
      }
      if (outputFLatency && ((outputFLatency * 100.0) >= 5.0)) {
        sprintf(tempStr, "\"%s - Secs Between Each SCNL Packet Group\"=%.3f;",
                hostName, outputFLatency); // {{ Average FL }}
        strcat(outStr, tempStr);
      }
      if (time_since_last && time_since_last > 0.0) {
        sprintf(tempStr, "\"%s - Secs Since Last Packet Received\"=%.3f;",
                hostName, time_since_last);
        strcat(outStr, tempStr);
      }
      if (outputDLatency && ((outputDLatency * 100.0) >= 5.0)) {
        sprintf(tempStr, "\"%s - Secs of Transport Latency\"=%.3f", hostName,
                outputDLatency); // {{ Average DL }}
        strcat(outStr, tempStr);
      }
      sprintf(tempStr, "\n");
      strcat(outStr, tempStr);

      logit("", "%s", outStr);

      int sockReturn;
      sockReturn = client_send_cmd(sock, outStr);

      if (sockReturn != OK) {
        logit("e", "Error: client_send_cmd(%d, %s) return = %d\n", sock, outStr,
              sockReturn);
        logit("e", "ERROR code on send: %d\n", smerr);
        exit(0);
      }

      if (debug) {
        logit("", "sendReturn = %d\n", sockReturn);
      }

      /////////////////////////////////////////////////////////////////////////////////////////////////////////

      sleptStatus[j] = 0;

      // Increment the reference time
      refTime[j] += increment;

      // If we do not have data to fill in intermediate times,
      // shift the reference time that we are looking for.
      if (refTime[j] < (inputTime[j][0] + t_offset)) {
        refTime[j] = (int)((inputTime[j][0] + t_offset) + increment);
      }
    } // end while (processing a time window)
  }   // end for (each tag)
  return 0;
}

/************************************************************************************/
