/*event_mode
 * pick_GM.c
 * Modified from pick_eew
 */

#define PICK_GM_VERSION "1.2.2 - 2024-06-11"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <time_ew.h>
#include <earthworm.h>
#include <transport.h>
#include <trace_buf.h>
#include <swap.h>
#include <trheadconv.h>
#include <chron3.h>
#include <math.h>
#include <limits.h>
#include "exportfilter.h"
#include "pick_gm.h"

#if defined(WIN32) || defined(WIN64)
	// Copied from linux libc sys/stat.h:
	#define S_ISREG(m) (((m) & S_IFMT) == S_IFREG)
	#define S_ISDIR(m) (((m) & S_IFMT) == S_IFDIR)
#endif

const char kPathSeparator =
#if defined(WIN32) || defined(WIN64)
							'\\';
#else
							'/';
#endif


/* Function prototypes
   *******************/
int  GetConfig( char *, GPARM * );
void LogConfig( GPARM * );
int  GetStaList( STATION **, int *, GPARM *, bool );
void LogStaList( STATION *, int );
int  CompareSCNL( const void *, const void * );
int  CompareSNL( const void *, const void * );
int  Restart( STATION *, GPARM *, int, int );
void Interpolate( STATION *, char *, int );
int  GetEwh( EWH * );
void reportGM(STATION* , GPARM* , EWH*  );
void reportGMX(STATION* , GPARM* , EWH*  );
void reportPSA(STATION*, GPARM*);
void playDummySniff(STATION *Sta, GPARM *Gparm, int nsamp, int32_t *TraceLong);

int sniff_gm( double, int, STATION *, int);
int sniff_gm2( double, int, STATION *, int);
int sniff_gmx(STATION *, STATION *);
int sniff_gm_dummy(int, STATION *);
int sniff_gm2_dummy(int, STATION *);

int writeGM(STATION *, GPARM *, EWH *, STATION *, int);
int writeGMX(STATION *, GPARM *, EWH *); 
int getPSA(STATION *Sta, GPARM *Gparm, double pkt_time, int is_last);

void Interpolate_simple(char *WaveBuf, int inter_sidx, int32_t inter_ssmp, int inter_eidx, int32_t inter_esmp, int nInter);
int anomalyFilter(char *FilteredTraceBuf);

// xmlthrd.c
thr_ret XmlThreadGM(void* ptr);
thr_ret XmlThreadGMX(void* ptr);

// spectrum.c
void response_spectrum(double *inacc, int npts, double *peroids, int nper, double dt, double damping, double *psa, STATION *Sta, GPARM *Gparm, int method);
typedef enum {
	METHOD_FFT = 0
	, METHOD_NEWMARK_BETA
	, METHOD_NIGAM_JENNINGS
} spectrum_method_t;


/***********************************************************
*              The main program starts here.              *
*                                                         *
*  Arguments:                                             *
*     argv[1] = Name of picker configuration file         *
*     argv[2] = (Optional) Tracebuffer file for offline   *
*                          mode                           *
***********************************************************/

int offline_mode;  /* Global flag to determine whether we work in offline mode or not*/
MSG_LOGO  fooMsg;               /* logo of retrieved message     */

int main (int argc, char **argv)
{
	int           i;                /* Loop counter */
	STATION       *StaArray = NULL; /* Station array */
	STATION       *StaGmxArray = NULL;	/* Station GMX array */
	char          *TraceBuf;        /* Pointer to waveform buffer */
	TRACE_HEADER  *TraceHead;       /* Pointer to trace header w/o loc code */
	TRACE2_HEADER *Trace2Head;      /* Pointer to header with loc code */
	int           *TraceLong;       /* Long pointer to waveform data */
	short         *TraceShort;      /* Short pointer to waveform data */
	long          MsgLen;           /* Size of retrieved message */
	MSG_LOGO      logo;             /* Logo of retrieved msg */
	MSG_LOGO      hrtlogo;          /* Logo of outgoing heartbeats */
	int           Nsta = 0;         /* Number of stations in list */
	int           NstaGmx = 0;
	time_t        then;             /* Previous heartbeat time */
	long          InBufl;           /* Maximum message size in bytes */
	GPARM         Gparm;            /* Configuration file parameters */
	EWH           Ewh;              /* Parameters from earthworm.h */
	char          *configfile;      /* Pointer to name of config file */
	pid_t         myPid;            /* Process id of this process */

	char    	type[3];
	STATION 	key;              	/* Key for binary search */
	STATION 	*Sta;             	/* Pointer to the station being processed */
	STATION 	*StaGmx;            /* Pointer to the station being processed */
	int			rc;               	/* Return code from tport_copyfrom() */
	time_t  	now;              	/* Current time */
	double  	GapSizeD;         	/* Number of missing samples (double) */
	int  		GapSize;          	/* Number of missing samples (integer) */
	unsigned char 	seq;        	/* msg sequence number from tport_copyfrom() */

	char	*tracename = NULL;
	FILE 	*tracefile = NULL;
	int 	loop_condition;

	/* Filtered */
	char          *FilteredTraceBuf;        /* Pointer to waveform buffer */
	TRACE2_HEADER *FilteredTrace2Head;      /* Pointer to header with loc code */
	int32_t       *FilteredTraceLong;       /* Long pointer to waveform data */
	short         *FilteredTraceShort;      /* Short pointer to waveform data */
	long			filteredSize;  /* size of message after user-filtering  */
	unsigned char	filteredType;  /* type of message after filtering       */

	unsigned  tidThrdGM;                /* Filter thread id              */
	unsigned  tidThrdGMX;                /* Filter thread id              */
	int numOfElementsInQueue;

	// 2018/12/04, jman, flag for using scnfilter
	int do_scnfilter = 0;

	// 2020/03/10, jman
	int32_t *buf;
	int32_t *sample;

	double pkt_time;
	long   maxTrace;
	
	struct timeval tv1, tv2;

	// set environment forcely, only for Visual Studio
#if defined(WIN32) || defined(WIN64)
	//_putenv("EW_DATA=d:\\earthworm\\korea\\data\\");
	//_putenv("EW_DATA_DIR=d:\\earthworm\\korea\\data\\");
	//_putenv("EW_HOME=d:\\earthworm");
	//_putenv("EW_INSTALLATION=INST_UNKNOWN");
	//_putenv("EW_LOG=d:\\earthworm\\korea\\log\\");
	//_putenv("EW_PARAMS=d:\\earthworm\\korea\\params");
	//_putenv("EW_VERSION=earthworm_7.9");
	//_putenv("TZ=GMT");
	//_putenv("SYS_NAME=TEMP");
#endif

	/* Check command line arguments */
	if (argc < 2 || (argc > 2 && argc != 3 && argc != 8)) {
		fprintf(stderr, "Usage: pick_gm <configfile> [tank_file]\n");
		fprintf(stderr, "       This module can be used as a standalone program (offline mode)\n");
		fprintf(stderr, "       by specifing the path to a tank file as second argument.\n");
		fprintf(stderr, " Version: %s\n", PICK_GM_VERSION);
		return -1;
	}
	configfile = argv[1];

	time(&now);

	offline_mode = 0;

	if (argc >= 3) {
		offline_mode = 1;
		tracename = argv[2];

		if ( (tracefile = fopen (tracename, "rb")) == NULL ) {
			perror (tracename);
			exit (-1);
		}
	}

	/* Initialize name of log-file & open it */
	logit_init (configfile, 0, 256, 1);

	if (offline_mode) {
		logit("", "Input file(tank): %s\n", tracename);
	}

	/* Get parameters from the configuration files */
	if ( GetConfig (configfile, &Gparm) == -1 ) {
		logit( "e", "pick_gm: GetConfig() failed. Exiting.\n" );
		return -1;
	}

	/* Look up info in the earthworm.h tables */
	if ( GetEwh( &Ewh ) < 0 ) {
		logit( "e", "pick_gm: GetEwh() failed. Exiting.\n" );
		return -1;
	}

	/* Specify logos of incoming waveforms and outgoing heartbeats */
	if( Gparm.nGetLogo == 0 ) {
		Gparm.nGetLogo = 2;
		Gparm.GetLogo  = (MSG_LOGO *) calloc( Gparm.nGetLogo, sizeof(MSG_LOGO) );
		if( Gparm.GetLogo == NULL ) {
			logit( "e", "pick_gm: Error allocating space for GetLogo. Exiting\n" );
			return -1;
		}
		Gparm.GetLogo[0].instid = Ewh.InstIdWild;
		Gparm.GetLogo[0].mod    = Ewh.ModIdWild;
		Gparm.GetLogo[0].type   = Ewh.TypeTracebuf2;

		Gparm.GetLogo[1].instid = Ewh.InstIdWild;
		Gparm.GetLogo[1].mod    = Ewh.ModIdWild;
		Gparm.GetLogo[1].type   = Ewh.TypeTracebuf;
	}

	hrtlogo.instid = Ewh.MyInstId;
	hrtlogo.mod    = Gparm.MyModId;
	hrtlogo.type   = Ewh.TypeHeartBeat;

	/* Set logo type for offlinemode
	*************************************/
	logo.type = Ewh.TypeTracebuf2;

	/* Get our own pid for restart purposes
	*************************************/
	myPid = getpid();
	if ( myPid == -1 ) {
		logit( "e", "pick_gm: Can't get my pid. Exiting.\n" );
		free( Gparm.GetLogo );
		free( Gparm.StaFile );
		return -1;
	}

	/* Initialize export filter
	*************************************/
	if( exportfilter_init() != 0 )
	{
		logit("e", "pick_gm: Scnfilter parameter was not set. Scnfilter will not be used.\n");
		do_scnfilter = 0;
	} else {
		do_scnfilter = 1;
	}

	/* Log the configuration parameters
	*************************************/
	LogConfig (&Gparm);

	/* Allocate the waveform buffer
	*************************************/
	InBufl = MAX_TRACEBUF_SIZ*2 + sizeof(int)*(Gparm.MaxGap-1);
	TraceBuf = (char *) malloc( (size_t) InBufl );
	if ( TraceBuf == NULL ) {
		logit( "et", "pick_gm: Cannot allocate waveform buffer\n" );
		free( Gparm.GetLogo );
		free( Gparm.StaFile );
		return -1;
	}
	buf = (int32_t *)malloc((size_t)InBufl);
	if (buf == NULL) {
		logit("et", "pick_gm: Cannot allocate waveform buffer\n");
		free(Gparm.GetLogo);
		free(Gparm.StaFile);
		free(TraceBuf);
		return -1;
	}
	sample = (int32_t *)malloc((size_t)InBufl);
	if (sample == NULL) {
		logit("et", "pick_gm: Cannot allocate waveform buffer\n");
		free(Gparm.GetLogo);
		free(Gparm.StaFile);
		free(TraceBuf);
		free(buf);
		return -1;
	}
	if (do_scnfilter)
	{
		FilteredTraceBuf = (char *) malloc( (size_t) InBufl );
		if ( FilteredTraceBuf == NULL ) {
			logit( "et", "pick_gm: Cannot allocate waveform buffer(FilteredTraceBuf)\n" );
			free( Gparm.GetLogo );
			free( Gparm.StaFile );
			free(TraceBuf);
			free(buf);
			free(sample);
			return -1;
		}
	} else {
		FilteredTraceBuf = TraceBuf;
	}

	/* Point to header and data portions of waveform message
	********************************************************/
	TraceHead  = (TRACE_HEADER *)TraceBuf;
	Trace2Head = (TRACE2_HEADER *)TraceBuf;
	TraceLong  = (int *) (TraceBuf + sizeof(TRACE_HEADER));
	TraceShort = (short *) (TraceBuf + sizeof(TRACE_HEADER));

	FilteredTrace2Head = (TRACE2_HEADER *)FilteredTraceBuf;
	FilteredTraceLong  = (int32_t *) (FilteredTraceBuf + sizeof(TRACE_HEADER));
	FilteredTraceShort = (short *) (FilteredTraceBuf + sizeof(TRACE_HEADER));

	/* Read the station list and return the number of stations found.
		Allocate the station list array
	*************************************/
	if ( GetStaList( &StaArray, &Nsta, &Gparm, true ) == -1 ) {
		logit( "e", "pick_gm: GetStaList() failed. Exiting.\n" );
		free( Gparm.GetLogo );
		free( Gparm.StaFile );
		free(TraceBuf);
		free(buf);
		free(sample);
		free( StaArray );
		return -1;
	}

	/* Input acceleration buffer for PSA
	*************************************/
	maxTrace = 0;
	for (i = 0; i < Nsta; i++) {
		StaArray[i].maxcnt = (int)StaArray[i].srate * (Gparm.SamplePeriod + 1); /* 1 minute + 60 seconds */
		StaArray[i].inacc = (double *)malloc(sizeof(double) * StaArray[i].maxcnt);
		if (StaArray[i].inst == 2) {
			StaArray[i].invel = (double *)malloc(sizeof(double) * StaArray[i].maxcnt);
		}
		else StaArray[i].invel = NULL;

		if (maxTrace < StaArray[i].maxcnt) maxTrace = StaArray[i].maxcnt;

	}

	logit("", "pick_gm: staArray len: %d\n", Nsta);

	if (maxTrace < Gparm.maxTrace) maxTrace = Gparm.maxTrace;
	logit("", "pick_gm: maxTrace - [%ld]\n", maxTrace);

	if (Gparm.WriteGMX) {
		if (GetStaList(&StaGmxArray, &NstaGmx, &Gparm, false) == -1)
		{
			logit("e", "pick_gm: GetStaList() for GmxArray failed. Exiting.\n");
			free(Gparm.GetLogo);
			free(Gparm.StaFile);
			free(TraceBuf);
			free(buf);
			free(sample);
			free(StaGmxArray);
			return -1;
		}
		logit("", "pick_gm: staGmxArray len: %d\n", NstaGmx);
	}

	if (Nsta == 0 || (Gparm.WriteGMX && NstaGmx == 0)) {
		logit( "et", "pick_gm: Empty station list(s). Exiting." );
		free( Gparm.GetLogo );
		free( Gparm.StaFile );
		free(TraceBuf);
		free(buf);
		free(sample);
		free( StaArray );
		return -1;
	}

	/* Sort the station list by SCNL */
	qsort (StaArray, Nsta, sizeof(STATION), CompareSCNL);
	if (Gparm.WriteGMX) {
		qsort(StaGmxArray, NstaGmx, sizeof(STATION), CompareSCNL);
	}

	/* Log the station list */
	LogStaList( StaArray, Nsta );

	/* Iint Earthworm enviroment 
	*****************************/
	if (!offline_mode) { /* if not in offline mode */
		/* Attach to existing transport rings */
		if ( Gparm.OutKey != Gparm.InKey ) {
			tport_attach( &Gparm.InRegion,  Gparm.InKey );
			tport_attach( &Gparm.OutRegion, Gparm.OutKey );
		} else {
			tport_attach( &Gparm.InRegion, Gparm.InKey );
			Gparm.OutRegion = Gparm.InRegion;
		}

		/* Flush the input ring */
		while ( tport_copyfrom( &Gparm.InRegion, Gparm.GetLogo, (short)Gparm.nGetLogo, 
						 &logo, &MsgLen, TraceBuf, MAX_TRACEBUF_SIZ, &seq) != 
						 GET_NONE );

		/* Get the time when we start reading messages.
		   This is for issuing heartbeats. */
		time( &then );
	}

	/* Create MsgQueue mutex */
	CreateMutex_ew();

	/* Allocate the message Queue */
	initqueue (&(Gparm.MsgQueueGM), 2000, 300);
	if (Gparm.WriteGMX) {
		initqueue(&(Gparm.MsgQueueGMX), 2000, 300);
	}

	/* Start decimator thread which will read messages from   *
	 * the Queue, fir them and write them to the OutRing */
	if (StartThreadWithArg (XmlThreadGM, (void *) &Gparm, (unsigned) THREAD_STACK, &tidThrdGM) == -1)
	{
		logit( "e", "pick_gm: Error starting XmlThreadGM thread.  Exitting.\n");
		exit (EW_FAILURE);
	}
	if (Gparm.WriteGMX) {
		if (StartThreadWithArg(XmlThreadGMX, (void *)&Gparm, (unsigned)THREAD_STACK, &tidThrdGMX) == -1)
		{
			logit("e", "pick_gm: Error starting XmlThreadGMX thread.  Exitting.\n");
			exit(EW_FAILURE);
		}
	}

	/* The loop condition depends on whether we are in offline_mode or not*/
	loop_condition=1;

	// 20171204, jman, Consider the case where the waveform does not come in
	time (&now);
	then = now - Gparm.HeartbeatInt - 1;

	// start time
	gettimeofday(&tv1, NULL);

	while (loop_condition) {

		if (offline_mode) {
			// TODO: check if the file is a valid trace buffer
			loop_condition = (fread (Trace2Head, sizeof(TRACE2_HEADER), 1, tracefile) > 0) &&
					 (fread (TraceLong, TraceHead->nsamp*sizeof(int), 1, tracefile) > 0);
		} else {
			loop_condition = tport_getflag( &Gparm.InRegion ) != TERMINATE &&
					 tport_getflag( &Gparm.InRegion ) != myPid;
		}

		if (!loop_condition) {
			logit("", "pick_gm: loop_condition break\n");
			break;
		}

		if (!offline_mode) {

			/* Get tracebuf or tracebuf2 message from ring */
			rc = tport_copyfrom( &Gparm.InRegion, Gparm.GetLogo, (short)Gparm.nGetLogo, 
						 &logo, &MsgLen, TraceBuf, MAX_TRACEBUF_SIZ, &seq );

			if ( rc == GET_NONE ) {

				// 20171204, jman, Consider waveform not inputed
				/* Send a heartbeat to the transport ring */
				time (&now);
				if ( (now - then) >= Gparm.HeartbeatInt ) {
					int  lineLen;
					char line[40];
					then = now;
					sprintf (line, "%ld %d\n", (long) now, (int) myPid);
					lineLen = strlen (line);
					if ( tport_putmsg (&Gparm.OutRegion, &hrtlogo, lineLen, line) !=
						PUT_OK ) {
						logit ("et", "pick_gm: Error sending heartbeat. Exiting.");
						break;
					}
				}

				sleep_ew( 100 );
				continue;
			}

			if ( rc == GET_NOTRACK )
				logit( "et", "pick_gm: Tracking error (NTRACK_GET exceeded)\n");

			if ( rc == GET_MISS_LAPPED )
				logit( "et", "pick_gm: Missed msgs (lapped on ring) "
					"before i:%d m:%d t:%d seq:%d\n",
					(int)logo.instid, (int)logo.mod, (int)logo.type, (int)seq );

			if ( rc == GET_MISS_SEQGAP )
				logit( "et", "pick_gm: Gap in sequence# before i:%d m:%d t:%d seq:%d\n",
					(int)logo.instid, (int)logo.mod, (int)logo.type, (int)seq );

			if ( rc == GET_TOOBIG ) {
				logit( "et", "pick_gm: Retrieved msg is too big: i:%d m:%d t:%d len:%d\n",
					(int)logo.instid, (int)logo.mod, (int)logo.type, MsgLen );
				continue;
			}

			/* If necessary, swap bytes in tracebuf message */
			if ( logo.type == Ewh.TypeTracebuf ) {
				if ( WaveMsgMakeLocal (TraceHead) < 0 ) {
					logit( "et", "pick_gm: WaveMsgMakeLocal() error.\n" );
					continue;
				}
			} else {
				if ( WaveMsg2MakeLocal (Trace2Head) < 0 ) {
					logit( "et", "pick_gm: WaveMsg2MakeLocal error.\n" );
					continue;
				}
			}

			/* Convert TYPE_TRACEBUF messages to TYPE_TRACEBUF2 */
			if ( logo.type == Ewh.TypeTracebuf )
				Trace2Head = TrHeadConv (TraceHead);
		}

		/* SCNL filter
		*************************************/
		if (do_scnfilter)
			if ( exportfilter( TraceBuf, MAX_TRACEBUF_SIZ, logo.type, &FilteredTraceBuf,
					&filteredSize, &filteredType ) == 0 )  continue;

		/* Look up SCNL number in the station list */
		{
			int j;
			for ( j = 0; j < 5; j++ ) key.sta[j]  = FilteredTrace2Head->sta[j];
			key.sta[5] = '\0';
			for ( j = 0; j < 3; j++ ) key.chan[j] = FilteredTrace2Head->chan[j];
			key.chan[3] = '\0';
			for ( j = 0; j < 2; j++ ) key.net[j]  = FilteredTrace2Head->net[j];
			key.net[2] = '\0';
			for ( j = 0; j < 2; j++ ) key.loc[j]  = FilteredTrace2Head->loc[j];
			key.loc[2] = '\0';
		}

		Sta = (STATION *) bsearch (&key, StaArray, Nsta, sizeof(STATION),
						CompareSCNL);
		if ( Sta == NULL )      /* SCNL not found */
			continue;

		/* If the samples are shorts, make them longs */
		strcpy(type, FilteredTrace2Head->datatype);
		if ((strcmp(type, "i2") == 0) || (strcmp(type, "s2") == 0)) {
			for (i = FilteredTrace2Head->nsamp - 1; i > -1; i--)
				FilteredTraceLong[i] = (int)FilteredTraceShort[i];
		}

		/* Check abnormal signal */
		/* It changes original values */
		if (anomalyFilter(FilteredTraceBuf)==EW_FAILURE) continue;

		/* Do this the first time we get a message with this SCNL */
		if ( Sta->first == 1 ) {
			Sta->endtime = FilteredTrace2Head->endtime;
			Sta->first = 0;

			playDummySniff(Sta, &Gparm, FilteredTrace2Head->nsamp, FilteredTraceLong);
		}

		/* If the samples are shorts, make them longs */
		strcpy(type, Trace2Head->datatype);
		if ((strcmp(type, "i2") == 0) || (strcmp(type, "s2") == 0)) {
			for (i = Trace2Head->nsamp - 1; i > -1; i--)
				TraceLong[i] = (int)TraceShort[i];
		}

		/* 
		 * Compute the number of samples since the end of the previous message.
		 * If (GapSize == 1), no data has been lost between messages.
		 * If (1 < GapSize <= Gparm.MaxGap), data will be interpolated.
		 * If (GapSize > Gparm.MaxGap), the picker will go into restart mode.
		*/
		GapSizeD = FilteredTrace2Head->samprate * (FilteredTrace2Head->starttime - Sta->endtime);

		if ( GapSizeD < 0. )          /* Invalid. Time going backwards. */
			GapSize = 0;
		else
		{
			// 2020/03/23, jman. Handling Exception.
			if (GapSizeD > (INT_MAX-1)) GapSize = INT_MAX;
			else GapSize = (int)(GapSizeD + 0.5);
		}

		/* Interpolate missing samples and prepend them to the current message */
		if ( (GapSize > 1) && (GapSize <= Gparm.MaxGap) )
			Interpolate (Sta, FilteredTraceBuf, GapSize);

		/* Announce large sample gaps */
		if ( GapSize > Gparm.MaxGap ) {
			int      lineLen;
			time_t   errTime;
			char     errmsg[80];
			MSG_LOGO logo;

			time( &errTime );
			sprintf( errmsg,
				"%ld %d Found %4d sample gap. Restarting channel %s.%s.%s.%s\n",
				(long) errTime, PK_RESTART, GapSize, Sta->sta, Sta->chan, Sta->net, Sta->loc );
			lineLen = strlen( errmsg );
			logo.type   = Ewh.TypeError;
			logo.mod    = Gparm.MyModId;
			logo.instid = Ewh.MyInstId;
			if(!offline_mode) tport_putmsg (&Gparm.OutRegion, &logo, lineLen, errmsg);
			else logit("et", "%s", errmsg);

		}

		if ( Restart (Sta, &Gparm, FilteredTrace2Head->nsamp, GapSize) ) 
		{
			/* Do this the first time we get a message with this SCNL */
			if (Sta->first == 1) {
				Sta->endtime = FilteredTrace2Head->endtime;
				Sta->first = 0;

				playDummySniff(Sta, &Gparm, FilteredTrace2Head->nsamp, FilteredTraceLong);
			}

			for ( i = 0; i < FilteredTrace2Head->nsamp; i++ )
			{
				pkt_time = (1.0*i)/(FilteredTrace2Head->samprate)+(FilteredTrace2Head->starttime);

				if (Sta->first_tm < 0) {
					Sta->first_tm = pkt_time + Gparm.StableTime;
				}

				if (pkt_time <= Sta->last_tm) continue;
				else Sta->last_tm = pkt_time;

				if (FilteredTraceLong[i] > Gparm.ClipLimit || FilteredTraceLong[i] < -Gparm.ClipLimit) {
					if (i == 0) FilteredTraceLong[i] = Sta->last_trace;
					else        FilteredTraceLong[i] = FilteredTraceLong[i - 1];
				}
				Sta->last_trace = TraceLong[i];

				switch (Gparm.sniffGmStyle)
				{
				case 0:
					sniff_gm(pkt_time, FilteredTraceLong[i], Sta, 1);
					break;
				case 1:
					sniff_gm2(pkt_time, FilteredTraceLong[i], Sta, 1);
					break;
				default:
					sniff_gm(pkt_time, FilteredTraceLong[i], Sta, 1);
					break;
				}

				if (Sta->is_oversec)
				{   
					getPSA(Sta, &Gparm, pkt_time, 0);
					if (pkt_time > Sta->first_tm) writeGM(Sta, &Gparm, &Ewh, StaArray, Nsta);
					Sta->is_oversec = false;
								
					if (Gparm.WriteGMX) {
						StaGmx = (STATION *)bsearch(&key, StaGmxArray, NstaGmx, sizeof(STATION), CompareSNL);
						if (StaGmx != NULL)
						{
							sniff_gmx(Sta, StaGmx);
							if (StaGmx->is_oversec)
							{
								if (pkt_time > Sta->first_tm) writeGMX(StaGmx, &Gparm, &Ewh);
								StaGmx->is_oversec = false;
							}
						}
					}
				}

				if (Gparm.numPeriods > 0)
				{
					if (pkt_time > Sta->first_tm) {
						if (Sta->inst == 2) Sta->invel[Sta->npts] = Sta->v;
						Sta->inacc[Sta->npts++] = Sta->a;
						if (fabs(Sta->a) > Sta->winpga) Sta->winpga = fabs(Sta->a);
						if (fabs(Sta->v) > Sta->winpgv) Sta->winpgv = fabs(Sta->v);
					}
				}
			} // end of for
		} else {

			/* Call sniff_gm */
			for ( i = 0; i < FilteredTrace2Head->nsamp; i++ )
			{
				pkt_time = (1.0*i)/(FilteredTrace2Head->samprate)+(FilteredTrace2Head->starttime);

				if (Sta->first_tm < 0) {
					Sta->first_tm = pkt_time + Gparm.StableTime;
				}

				if (pkt_time <= Sta->last_tm) continue;
				else Sta->last_tm = pkt_time;

				if (FilteredTraceLong[i] > Gparm.ClipLimit || FilteredTraceLong[i] < -Gparm.ClipLimit) {
					if (i == 0) FilteredTraceLong[i] = 0;
					else        FilteredTraceLong[i] = FilteredTraceLong[i - 1];
				}
				
				switch (Gparm.sniffGmStyle)
				{
				case 0:
					sniff_gm(pkt_time, FilteredTraceLong[i], Sta, 0);
					break;
				case 1:
					sniff_gm2(pkt_time, FilteredTraceLong[i], Sta, 0);
					break;
				default:
					sniff_gm(pkt_time, FilteredTraceLong[i], Sta, 0);
					break;
				}

				if (Sta->is_oversec)
				{
					getPSA(Sta, &Gparm, pkt_time, 0);
					if (pkt_time > Sta->first_tm) writeGM(Sta, &Gparm, &Ewh, StaArray, Nsta);
					Sta->is_oversec = false;
					
					if (Gparm.WriteGMX) {

						StaGmx = (STATION *)bsearch(&key, StaGmxArray, NstaGmx, sizeof(STATION), CompareSNL);
						if (StaGmx != NULL)
						{
							sniff_gmx(Sta, StaGmx);
							if (StaGmx->is_oversec)
							{
								if (pkt_time > Sta->first_tm) writeGMX(StaGmx, &Gparm, &Ewh);
								StaGmx->is_oversec = false;
							}
						}
					}
				}

				if (Gparm.numPeriods > 0)
				{
					if (pkt_time > Sta->first_tm) {
						if (Sta->inst == 2) Sta->invel[Sta->npts] = Sta->v;
						Sta->inacc[Sta->npts++] = Sta->a;
						if (fabs(Sta->a) > Sta->winpga) Sta->winpga = fabs(Sta->a);
						if (fabs(Sta->v) > Sta->winpgv) Sta->winpgv = fabs(Sta->v);
					}				
				}
			} // end of for 
		}

		/* Save time and amplitude of the end of the current message */
		Sta->enddata = FilteredTraceLong[FilteredTrace2Head->nsamp - 1];
		Sta->endtime = FilteredTrace2Head->endtime;

		if (!offline_mode) {
			/* Send a heartbeat to the transport ring */
			time (&now);
			if ( (now - then) >= Gparm.HeartbeatInt ) {
				int  lineLen;
				char line[40];

				then = now;

				sprintf (line, "%ld %d\n", (long) now, (int) myPid);
				lineLen = strlen (line);
				if ( tport_putmsg (&Gparm.OutRegion, &hrtlogo, lineLen, line) !=
					PUT_OK ) {
					logit ("et", "pick_gm: Error sending heartbeat. Exiting.");
					break;
				}
			}
		}
	} /* End of loop on waveform messages */

	if (!offline_mode) { /* if not in offline_mode */
		/* Detach from the ring buffers */
		if ( Gparm.OutKey != Gparm.InKey ) {
			tport_detach( &Gparm.InRegion );
			tport_detach( &Gparm.OutRegion );
		} else {
			tport_detach( &Gparm.InRegion );
		}
		logit( "t", "Termination requested. Exiting.\n" );
	}

	if (Gparm.OutputByFile) {
		if ( Gparm.fpGM != NULL) 
		{
			/* Reflects the last changes */
			RequestMutex ();
			enqueue (&(Gparm.MsgQueueGM), Gparm.strFpGM, strlen(Gparm.strFpGM), fooMsg); 
			ReleaseMutex_ew ();
			fclose(Gparm.fpGM);
		}
		if ( Gparm.fpGMX != NULL) 
		{
			RequestMutex ();
			enqueue (&(Gparm.MsgQueueGMX), Gparm.strFpGMX, strlen(Gparm.strFpGMX), fooMsg); 
			ReleaseMutex_ew ();
			fclose(Gparm.fpGMX);
		}
	}

	/* waiting for thread to process the data */
	sleep_ew(200);
	if ( Gparm.OutputByFile ) 
	{
		while(true)
		{	
			numOfElementsInQueue = getNumOfElementsInQueue(&(Gparm.MsgQueueGM));
			if (numOfElementsInQueue > 0 || !Gparm.isCompleteThrdGM ) {
				sleep_ew(200);
				continue;
			}
			printf("ThreadGM end..\n");

			if (Gparm.WriteGMX)
			{
				numOfElementsInQueue = getNumOfElementsInQueue(&(Gparm.MsgQueueGMX));
				if (numOfElementsInQueue > 0 || !Gparm.isCompleteThrdGMX) {
					sleep_ew(200);
					continue;
				}
				printf("ThreadGMX end..\n");
			}		

			sleep_ew(200);
			break;
		}
	}

	// end time
	gettimeofday(&tv2, NULL);

	logit("ot", "Takes time....  %f\n",
		(double)(tv2.tv_usec - tv1.tv_usec) / 1000000 +
		(double)(tv2.tv_sec - tv1.tv_sec)
	);

	free (Gparm.GetLogo);
	free (Gparm.StaFile);
	free (StaArray);
	free (TraceBuf);

	return 0;
}

/**************************************************************************
* playDummySniff()  Make dummy data before waveform's starting point
**************************************************************************/
void playDummySniff(STATION *Sta, GPARM *Gparm, int nsamp, int32_t *TraceLong)
{
	int targetLength = Sta->srate * Gparm->dummyLeadSec;		 // 10sec
	int duplicateCnt = ceil(targetLength / nsamp);
	int z, y;

	if (Gparm->dummyLeadSec == 0) return;

	for (z = nsamp - 1; z >= 0; z--)
	{
		for (y = 0; y < duplicateCnt; y++)
		{
			if (!Gparm->sniffGmStyle) sniff_gm_dummy(TraceLong[z], Sta);
			else sniff_gm2_dummy(TraceLong[z], Sta);
		}
	}
}

/**************************************************************************
* dateparser()  Convert to year, mon, day, hour, min. sec. nsec from epoch
**************************************************************************/
void dateparser(double t, int *year, int *mon, int *day, int *hour, int *min, int *sec, int *nsec)
{
	time_t    tt;       /* time as time_t                  */
	struct tm stm;      /* time as struct tm               */
	int       t_hsec;   /* hundredths-seconds part of time */

						/* Convert double time to other formats
						**************************************/
	t += 0.005;  /* prepare to round to the nearest 100th */
	tt = (time_t)t;
	t_hsec = (int)((t - tt) * 100.);
	gmtime_ew(&tt, &stm);

	*year = stm.tm_year + 1900;
	*mon = stm.tm_mon + 1;
	*day = stm.tm_mday;
	*hour = stm.tm_hour;
	*min = stm.tm_min;
	*sec = stm.tm_sec;
	*nsec = t_hsec;

	return;
}

int writeGM(STATION *Sta, GPARM *Gparm, EWH *Ewh, STATION *StaArray, int Nsta)
{
	struct Greg g;
	struct stat st = {0};
	double  sec1970 = GSEC1970;
	char year[5];
	char mmdd[5];
	char hour[3];
	char minute[3];
	char temp[300];
	int ret;
	int fyr, fmo, fdy, fhr, fmn, fsec, fnsec;

	if (Gparm->OutputByFile) {

		dateparser(Sta->tm, &fyr, &fmo, &fdy, &fhr, &fmn, &fsec, &fnsec);
		sprintf(year, "%4d", fyr);
		sprintf(mmdd, "%02d%02d", fmo, fdy);
		sprintf(hour, "%02d", fhr);
		sprintf(minute, "%02d", fmn);

		// check year
		sprintf(temp, "%s%c%s", Gparm->OutputFolder, kPathSeparator, year);

		//if (stat(temp, &st) == 0 && S_ISDIR(st.st_mode)) {}
		//else 
		//{
		#if defined(WIN32) || defined(WIN64)
		_mkdir(temp);
		#else
		mkdir(temp, 0755);
		#endif

		// check mmdd
		sprintf(temp, "%s%c%s%c%s", Gparm->OutputFolder, kPathSeparator, year, kPathSeparator, mmdd);
		//if (stat(temp, &st) == 0 && S_ISDIR(st.st_mode)) {}
		//else 
		//{
		#if defined(WIN32) || defined(WIN64)
		_mkdir(temp);
		#else
		mkdir(temp, 0755);
		#endif

		// check hour
		sprintf(temp, "%s%c%s%c%s%c%s", Gparm->OutputFolder, kPathSeparator, year, kPathSeparator, mmdd, kPathSeparator, hour);
		//if (stat(temp, &st) == 0 && S_ISDIR(st.st_mode)) {}
		//else 
		//{
		#if defined(WIN32) || defined(WIN64)
		_mkdir(temp);
		#else
		mkdir(temp, 0755);
		#endif

		// check fopen
		sprintf(temp, "%s%c%s%c%s%c%s%c%s%s%s%s.gm", Gparm->OutputFolder, kPathSeparator, year, kPathSeparator, mmdd, kPathSeparator, hour, kPathSeparator
			, year, mmdd, hour, minute);
		if ( Gparm->fpGM == NULL) {
			strcpy(Gparm->strFpGM, temp);
			Gparm->fpGM = fopen(temp,"a");
			if( Gparm->fpGM == NULL )
			{
				printf("File open error %s !\n", temp);
				return 0;
			}
		} else {
			if (!strcmp(temp, Gparm->strFpGM)) {
				// same
			} 
			else 
			{
				// different
				fclose(Gparm->fpGM);

				RequestMutex ();
				ret = enqueue (&(Gparm->MsgQueueGM), Gparm->strFpGM, strlen(Gparm->strFpGM), fooMsg); 
				ReleaseMutex_ew ();
				sleep_ew(1);

				strcpy(Gparm->strFpGM, temp);
				Gparm->fpGM = fopen(temp,"a");
				if( Gparm->fpGM == NULL )
				{
					printf("File open error %s !!\n", temp);
					return 0;
				}
			}
		}

	}	
	reportGM(Sta, Gparm, Ewh);

	return 1;
}

int writeGMX(STATION *Sta, GPARM *Gparm, EWH *Ewh) 
{
	//struct Greg g;
	struct stat st = {0};
	double  sec1970 = GSEC1970;
	char year[5];
	char mmdd[5];
	char hour[3];
	char minute[3];
	char temp[300];
	int ret;
	int fyr, fmo, fdy, fhr, fmn, fsec, fnsec;

	if (Gparm->OutputByFile) {

		dateparser(Sta->tm, &fyr, &fmo, &fdy, &fhr, &fmn, &fsec, &fnsec);
		sprintf(year, "%4d", fyr);
		sprintf(mmdd, "%02d%02d", fmo, fdy);
		sprintf(hour, "%02d", fhr);
		sprintf(minute, "%02d", fmn);

		// check year
		sprintf(temp, "%s%c%s", Gparm->OutputFolder, kPathSeparator, year);
		//if (stat(temp, &st) == 0 && S_ISDIR(st.st_mode)) {}
		//else 
		//{
			mkdir(temp, 0755);
			//sprintf(systemCmd, "[ -d %s ] || mkdir %s", temp, temp);
			//system(systemCmd);
		//}

		// check mmdd
		sprintf(temp, "%s%c%s%c%s", Gparm->OutputFolder, kPathSeparator, year, kPathSeparator, mmdd);
		//if (stat(temp, &st) == 0 && S_ISDIR(st.st_mode)) {}
		//else 
		//{
			mkdir(temp, 0755);
			//sprintf(systemCmd, "[ -d %s ] || mkdir %s", temp, temp);
			//system(systemCmd);
		//}

		// check hour
		sprintf(temp, "%s%c%s%c%s%c%s", Gparm->OutputFolder, kPathSeparator, year, kPathSeparator, mmdd, kPathSeparator, hour);
		//if (stat(temp, &st) == 0 && S_ISDIR(st.st_mode)) {}
		//else 
		//{
			mkdir(temp, 0755);
			//sprintf(systemCmd, "[ -d %s ] || mkdir %s", temp, temp);
			//system(systemCmd);
		//}

		// check fopen
		sprintf(temp, "%s%c%s%c%s%c%s%c%s%s%s%s.gmx", Gparm->OutputFolder, kPathSeparator, year, kPathSeparator, mmdd, kPathSeparator, hour, kPathSeparator
			, year, mmdd, hour, minute);
		if ( Gparm->fpGMX == NULL) {
			strcpy(Gparm->strFpGMX, temp);
			Gparm->fpGMX = fopen(temp,"a");
			if( Gparm->fpGMX == NULL )
			{
				printf("File open error %s !\n", temp);
				return 0;
			}
		} else {
			if (!strcmp(temp, Gparm->strFpGMX)) {
				// same
			} 
			else 
			{
				// different
				fclose(Gparm->fpGMX);

				RequestMutex ();
				ret = enqueue (&(Gparm->MsgQueueGMX), Gparm->strFpGMX, strlen(Gparm->strFpGMX), fooMsg); 
				ReleaseMutex_ew ();
				sleep_ew(1);

				strcpy(Gparm->strFpGMX, temp);
				Gparm->fpGMX = fopen(temp,"a");
				if( Gparm->fpGMX == NULL )
				{
					printf("File open error %s !\n", temp);
					return 0;
				}
			}
		}
	}
	
	reportGMX(Sta, Gparm, Ewh);
	return 1;
}


/*******************************************************
*                      GetEwh()                       *
*                                                     *
*      Get parameters from the earthworm.h file.      *
*******************************************************/
int GetEwh( EWH *Ewh )
{
	if ( GetLocalInst( &Ewh->MyInstId ) != 0 )
	{
		logit( "e", "pick_gm: Error getting MyInstId.\n" );
		return -1;
	}

	if ( GetInst( "INST_WILDCARD", &Ewh->InstIdWild ) != 0 )
	{
		logit( "e", "pick_gm: Error getting InstIdWild.\n" );
		return -2;
	}
	if ( GetModId( "MOD_WILDCARD", &Ewh->ModIdWild ) != 0 )
	{
		logit( "e", "pick_gm: Error getting ModIdWild.\n" );
		return -3;
	}
	if ( GetType( "TYPE_HEARTBEAT", &Ewh->TypeHeartBeat ) != 0 )
	{
		logit( "e", "pick_gm: Error getting TypeHeartbeat.\n" );
		return -4;
	}
	if ( GetType( "TYPE_ERROR", &Ewh->TypeError ) != 0 )
	{
		logit( "e", "pick_gm: Error getting TypeError.\n" );
		return -5;
	}
	if ( GetType( "TYPE_STRONG_MOTION_GM", &Ewh->TypeStrongMotionGM ) != 0 )
	{
		logit( "e", "pick_gm: Error getting TypeStrongMotionGM.\n" );
		return -6;
	}
	/*if ( GetType( "TYPE_STRONG_MOTION_GMX", &Ewh->TypeStrongMotionGMX ) != 0 )
	{
		logit( "e", "pick_gm: Error getting TypeStrongMotionGMX.\n" );
		return -7;
	}*/
	if ( GetType( "TYPE_TRACEBUF", &Ewh->TypeTracebuf ) != 0 )
	{
		logit( "e", "pick_gm: Error getting TYPE_TRACEBUF.\n" );
		return -8;
	}
	if ( GetType( "TYPE_TRACEBUF2", &Ewh->TypeTracebuf2 ) != 0 )
	{
		logit( "e", "pick_gm: Error getting TYPE_TRACEBUF2.\n" );
		return -9;
	}	
	return 0;
}


int anomalyFilter(char *TraceBuf)
{
	int inter_sidx, inter_eidx, inter_ssmp, inter_esmp, nInter, flag, i, j, zsamp;
	TRACE2_HEADER *header = (TRACE2_HEADER *)TraceBuf;
	int32_t *sample = (int *)(TraceBuf + sizeof(TRACE_HEADER));
	int nsamp = header->nsamp;
	double   SampleInterval = 1. / header->samprate;

	inter_sidx = inter_ssmp = inter_eidx = inter_esmp = nInter = zsamp = 0;
	
	for (i = 1; i < nsamp; i++)
	{
		double pkt_time = (1.0*i) / (header->samprate) + (header->starttime);

		// detect abnormal signal
		if (sample[i] == 0x10000000)
		{
			if (sample[i - 1] == 0x10000000)
			{
				// continuous status
			}
			else
			{
				// go to abnormal
				inter_sidx = i;
				inter_ssmp = sample[i - 1];
			}
		}
		else
		{
			if (sample[i - 1] == 0x10000000)
			{
				// back to normal
				inter_eidx = i - 1;
				inter_esmp = sample[i];
				nInter = inter_eidx - inter_sidx + 1;
			}
			else
			{
				// continuous status
			}
		}

		// interpolate
		if (inter_sidx > 0 && inter_eidx < nsamp - 1 && nInter > 0)
		{
			Interpolate_simple(TraceBuf, inter_sidx, inter_ssmp, inter_eidx, inter_esmp, nInter);
			//printf("[%f-%f] Packet interpolated.\n", header->starttime, header->endtime);
			//printf("Interpolate. inter_s: %d, inter_e: %d, nInter: %d, nsmp: %d", inter_sidx, inter_eidx, nInter, nsamp);
			logit("", "anomalyFilter: Found anomaly signal in part of the packet. Interpolate signal. packet: [%f-%f]\n", header->starttime, header->endtime);
			inter_sidx = inter_eidx = nInter = 0;
		}
	}

	// Check abnormal
	if (sample[nsamp - 1] == 0x10000000)
	{
		inter_eidx = nsamp - 1;
		inter_esmp = 0;
		nInter = inter_eidx - inter_sidx + 1;
	}

	// Adjust header or drop frame
	if (nInter > 0)
	{
		//printf("inter_s: %d, inter_e: %d, nInter: %d, nsmp: %d\n", inter_sidx, inter_eidx, nInter, nsamp);
		
		if (inter_sidx == 0 && inter_eidx == nsamp - 1)
		{
			// Drop packet
			logit("", "anomalyFilter: Found anomaly signal in whole packet. Drop packet. packet: [%f-%f]\n", header->starttime, header->endtime);
			return EW_FAILURE;
		}
		
		if (inter_sidx == 0)
		{
			// Adjust header starttime and nsamp
			// CASE. 2017_1012_164108, DACB.HGE.KS.--, PACKET: [1507826566.098400-1507826567.088400], ABNORMAL: [1507826566.098400-1507826566.958400]
			//printf("[%f-%f] Adjust start time and nsamp.\n", header->starttime, header->endtime);
			logit("", "anomalyFilter: Found anomaly signal in part of the packet. Change start time and nsamp. packet: [%f-%f]\n", header->starttime, header->endtime);
			for (j = 0; j < nsamp - nInter; j++)
			{
				sample[j] = sample[inter_eidx + j + 1];
			}
			header->nsamp = nsamp - nInter;
			header->starttime = header->starttime + SampleInterval * (inter_eidx + 1);
			
			return EW_SUCCESS;
		}
		
		if (inter_eidx == nsamp - 1)
		{
			// Adjust header endtime and nsamp
			// CASE. 2017_1012_164108, DACB.HGE.KS.--, PACKET: [1507826556.228400-1507826557.218400], ABNORMAL: [1507826556.968400-1507826557.218400]
			//printf("[%f-%f] Adjust end time and nsamp.\n", header->starttime, header->endtime);
			logit("", "anomalyFilter: Found anomaly signal in part of the packet. Change end time and nsamp. packet: [%f-%f]\n", header->starttime, header->endtime);
			header->nsamp = header->nsamp - nInter;
			header->endtime = header->starttime + SampleInterval * (inter_sidx -1);
			return EW_SUCCESS;
		}

	}
	
	//printf("[%f-%f] Packet passed\n", header->starttime, header->endtime);
	return EW_SUCCESS;
	
}


/* METHOD - NEWMARK_BETA/NIGAM_JENNINGS */
int getPSA(STATION *Sta, GPARM *Gparm, double pkt_time, int is_last)
{
	int    k;
	long   pkt_minute;
	double damping = 0.05;

	if (!Gparm->numPeriods) return (0);


	pkt_minute = (long)pkt_time / (Gparm->SamplePeriod / 2);
	
	for (k = 0; k < Gparm->numPeriods; k++) {
		Sta->psa[k] = 0.0;
	}

	if (Sta->pkt_minute == 0) {
		Sta->pkt_minute = pkt_minute;
	}
	else {

		if (Sta->npts > 2 && (is_last || pkt_minute > Sta->pkt_minute)) {
			response_spectrum(Sta->inacc, Sta->npts, Gparm->PsaPeriods, Gparm->numPeriods, Sta->dt, damping, Sta->psa, Sta, Gparm, Gparm->PsaMethod);

			int low_cut = (Gparm->SamplePeriod * Sta->srate) / 2;
			if (Sta->npts > low_cut) {
				memcpy(Sta->inacc, &Sta->inacc[Sta->npts - low_cut], sizeof(double)*low_cut);
				Sta->npts = low_cut;
			}

			Sta->pkt_minute = pkt_minute;
			Sta->winpga = 0;
			Sta->winpgv = 0;
			Sta->winpgd = 0;
		}
	}

	return (0);
}
