#include <math.h>
#include <stdio.h>
#include <string.h>
#include "sqlite3.h"
#define __ARCDB_C__
#include "arcdb.h"
#include "earthworm.h"

static char* arcdb_path;
static sqlite3 *db = NULL;
static int arcdb_flags = SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE;
static sqlite3_stmt *delete_stmt = NULL, *select_stmt = NULL, *insert_stmt = NULL, *update_stmt = NULL;

static void reset_statement(sqlite3_stmt* stmt) {
    if (stmt != NULL) {
        sqlite3_reset(stmt);
        sqlite3_clear_bindings(stmt);
    }
}

static void init_statements() {
    reset_statement(delete_stmt);
    reset_statement(select_stmt);
    reset_statement(insert_stmt);
    reset_statement(update_stmt);
}

static int open_if_needed() {
    int rc = 0;
    if (db == NULL) {
        rc = sqlite3_open_v2(arcdb_path, &db, arcdb_flags, NULL);
        if (rc != SQLITE_OK) {
            logit("et", "Can't open database %s (%i): %s\n", arcdb_path, rc, sqlite3_errmsg(db));
            arcdb_close();
        }
    }
    return rc;
}

/**
* Close the ARC database.
*/
void arcdb_close(void) {
    sqlite3_finalize(delete_stmt);
    sqlite3_finalize(select_stmt);
    sqlite3_finalize(insert_stmt);
    sqlite3_finalize(update_stmt);
    sqlite3_close(db);
    delete_stmt = NULL;
    select_stmt = NULL;
    insert_stmt = NULL;
    update_stmt = NULL;
    db = NULL;
}

static int arcdb_exec(char *sql, sqlite3_stmt **pstmt) {
    int rc = sqlite3_prepare_v2(db, sql, -1, pstmt, NULL);
    if(SQLITE_OK != rc) {
        logit("et", "Can't prepare statment %s (%i): %s\n", sql, rc, sqlite3_errmsg(db));
        sqlite3_finalize(*pstmt);
        return 4;
    }
    rc = sqlite3_step(*pstmt);
    if(SQLITE_DONE != rc) {
        logit("et", "Statement failed %s (%i): %s\n", sql, rc, sqlite3_errmsg(db));
        return 5;
    }
    return 0;
}

/**
* Initialiize amd then open a connection to the ARC database.
* @param path the path to the database file
* @return 0 if success, non-zero otherwise.
*/
int arcdb_init(char *path) {
    int count = 0;
    char *sql;
    sqlite3_stmt *stmt;
    int rc;
    arcdb_path = path;
    if ( (arcdb_flags & SQLITE_OPEN_CREATE) == 0 ) {
        return arcdb_open();
    }
    if ((rc = open_if_needed()) != SQLITE_OK) {
        return 1;
    }
    sql = "SELECT COUNT(name) FROM PRAGMA_TABLE_INFO('Events')";
    rc = sqlite3_prepare_v2(db, sql, -1, &stmt, NULL);
    if(SQLITE_OK != rc) {
        logit("et", "Can't prepare statment %s (%i): %s\n", sql, rc, sqlite3_errmsg(db));
        return 2;
    }
    count = 0;
    while(SQLITE_ROW == (rc = sqlite3_step(stmt))) {
        count = sqlite3_column_int(stmt, 0);
    }
    if(SQLITE_DONE != rc) {
        logit("et", "select statement didn't finish with DONE (%i): %s\n", rc, sqlite3_errmsg(db));
    }
    sqlite3_finalize(stmt);
    sql = NULL;
    switch (count) {
        case 0: // empty database
            logit("t","creating new Events table\n");
            sql = "CREATE TABLE Events(PdlID TEXT, TimeStamp TEXT, magnitude REAL, version INTEGER, update_count INTEGER, flags INTEGER)";
            break;
        case 2: // old database without magnitude, version, update_count and flags
            logit("t","adding magnitude, version, update_count and flags columns\n");
            sql = "ALTER TABLE Events ADD COLUMN magnitude REAL";
            if ((rc = arcdb_exec(sql, &stmt)) != 0) {
                return rc;
            }
            sql = "ALTER TABLE Events ADD COLUMN version INTEGER";
            if ((rc = arcdb_exec(sql, &stmt)) != 0) {
                return rc;
            }
            sql = "ALTER TABLE Events ADD COLUMN update_count INTEGER";
            if ((rc = arcdb_exec(sql, &stmt)) != 0) {
                return rc;
            }
            sql = "ALTER TABLE Events ADD COLUMN flags INTEGER";
            break;
        case 3: // old database without version, update_count and flags
            logit("t","adding version, update_count and flags column\n");
            sql = "ALTER TABLE Events ADD COLUMN version INTEGER";
            if ((rc = arcdb_exec(sql, &stmt)) != 0) {
                return rc;
            }
            sql = "ALTER TABLE Events ADD COLUMN update_count INTEGER";
            if ((rc = arcdb_exec(sql, &stmt)) != 0) {
                return rc;
            }
            sql = "ALTER TABLE Events ADD COLUMN flags INTEGER";
            break;
        case 4: // old database without update_count and flags
            logit("t","adding update_count and flags column\n");
            sql = "ALTER TABLE Events ADD COLUMN update_count INTEGER";
            if ((rc = arcdb_exec(sql, &stmt)) != 0) {
                return rc;
            }
            sql = "ALTER TABLE Events ADD COLUMN flags INTEGER";
            break;
        case 5: // old database without flags
            logit("t","adding flags column\n");
            sql = "ALTER TABLE Events ADD COLUMN flags INTEGER";
            break;
       case 6:
          break;
        default:
            logit("et", "Unexpected number of columns: %d\n", count);
            break;
    }
    if (sql != NULL && (rc = arcdb_exec(sql, &stmt)) != 0) {
        return rc;
    }
    return arcdb_open();
}

/**
* 0pen a connection to the ARC database.
* @return 0 if success, non-zero otherwise.
*/
int arcdb_open(void) {
    char *sql;
    int rc;
    if ((rc = open_if_needed()) != SQLITE_OK) {
        return 1;
    }
    if ( (arcdb_flags & SQLITE_OPEN_READWRITE) == 0) {
        sql = "SELECT ROWID, PdlID, magnitude, update_count, version, flags FROM Events WHERE ROWID = ? LIMIT 1";
        rc = sqlite3_prepare_v2(db, sql, -1, &select_stmt, NULL);
        if(SQLITE_OK != rc) {
            logit("et", "Can't prepare statment %s (%i): %s\n", sql, rc, sqlite3_errmsg(db));
            return 2;
        }
    } else {
        sql = "SELECT ROWID, PdlID, magnitude, update_count, version, flags FROM Events WHERE PdlID = ? LIMIT 1";
        rc = sqlite3_prepare_v2(db, sql, -1, &select_stmt, NULL);
        if(SQLITE_OK != rc) {
            logit("et", "Can't prepare statment %s (%i): %s\n", sql, rc, sqlite3_errmsg(db));
            return 3;
        }
        sql = "DELETE FROM Events WHERE TimeStamp < ?;";
        rc = sqlite3_prepare_v2(db, sql, -1, &delete_stmt, NULL);
        if(SQLITE_OK != rc) {
            logit("et", "Can't prepare statment %s (%i): %s\n", sql, rc, sqlite3_errmsg(db));
            return 4;
        }
        sql = "INSERT INTO Events(PdlID,TimeStamp,magnitude,version,update_count,flags) VALUES(?,datetime('now'),?,?,?,?);";
        rc = sqlite3_prepare_v2(db, sql, -1, &insert_stmt, NULL);
        if(SQLITE_OK != rc) {
            logit("et", "Can't prepare statment %s (%i): %s\n", sql, rc, sqlite3_errmsg(db));
            return 5;
        }
        sql = "UPDATE Events SET TimeStamp=date('now'),magnitude=?,version=?,update_count=?,flags=? WHERE PdlID=?;";
        rc = sqlite3_prepare_v2(db, sql, -1, &update_stmt, NULL);
        if(SQLITE_OK != rc) {
            logit("et", "Can't prepare statment %s (%i): %s\n", sql, rc, sqlite3_errmsg(db));
            return 6;
        }
    }
    return 0;
}

/**
 * Set the ARC database open flags.
 * @param flags the open flags or 0 for the default (reading and writing, and is created if it does not already exist)
 */
void arcdb_setflags(int flags) {
    if ( flags == 0 ) {
        flags = SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE;
    }
    arcdb_flags = flags;
}

/**
 * Shut down the ARC database.
 */
void arcdb_shutdown() {
    arcdb_close();
    sqlite3_shutdown();
}

/**
*  Delete events from the ARC dtabase.
*  @param dateStr the date string  to delete.
*/
void arcdb_delete_events(char *dateStr) {
    int rc;
    rc = sqlite3_bind_text(delete_stmt, 1, dateStr, strlen(dateStr), NULL);
    if(SQLITE_OK != rc) {
        logit("et", "Error binding value in statment (%i): %s\n", rc, sqlite3_errmsg(db));
        return;
    }
    rc = sqlite3_step(delete_stmt);
    if(SQLITE_DONE != rc) {
        logit("et", "delete statement didn't return DONE (%i): %s\n", rc, sqlite3_errmsg(db));
        return;
    }
}

static int do_get_event(ARCEVENTINFO *arcEventInfo) {
// char *eventID, long *rowid, double *old_mag, long *old_version, int *old_update_count) {
    int rc;
    if ( (arcdb_flags & SQLITE_OPEN_READWRITE) == 0) { // if db is read-only
       // select with rowid
       rc = sqlite3_bind_int64(select_stmt, 1, arcEventInfo->rowid);
    } else {
       // select with event ID
       rc = sqlite3_bind_text(select_stmt, 1, arcEventInfo->eventID,
                              strlen(arcEventInfo->eventID), NULL);
    }
    if(SQLITE_OK != rc) {
        logit("et", "Error binding value in statment (%i): %s\n", rc, sqlite3_errmsg(db));
        return 2;
    }
    rc = sqlite3_step(select_stmt);
    if(SQLITE_ROW == rc) {
        const char *str;
        int last;
        arcEventInfo->rowid = sqlite3_column_int(select_stmt, 0);
        str = (const char *) sqlite3_column_text(select_stmt, 1);
        last = sizeof(arcEventInfo->eventID) - 1;
        strncpy( arcEventInfo->eventID, str, last );
        arcEventInfo->eventID[last] = '\0';
        arcEventInfo->mag = sqlite3_column_double(select_stmt, 2);
        arcEventInfo->update_count = sqlite3_column_int(select_stmt, 3);
        arcEventInfo->version = sqlite3_column_int64(select_stmt, 4);
        arcEventInfo->flags = sqlite3_column_int(select_stmt, 5);
        rc = sqlite3_step(select_stmt);
    } else { // if no row
       ARCEVENTINFO_CLEAR(arcEventInfo);
    }
    if(SQLITE_DONE != rc) {
        logit("et", "select statement didn't return DONE (%i): %s\n", rc, sqlite3_errmsg(db));
        return 3;
    }
    return 0;
}

static int do_save_event(ARCEVENTINFO *arcEventInfo) {
    int rc;
    char *prefix;
    if(arcEventInfo->rowid != 0) {
        rc = sqlite3_bind_double(update_stmt, 1, arcEventInfo->mag);
        if(SQLITE_OK != rc) {
            logit("et", "Error binding value in statment (%i): %s\n", rc, sqlite3_errmsg(db));
            return 2;
        }
        rc = sqlite3_bind_int64(update_stmt, 2, arcEventInfo->version);
        if(SQLITE_OK != rc) {
            logit("et", "Error binding value in statment (%i): %s\n", rc, sqlite3_errmsg(db));
            return 3;
        }
        rc = sqlite3_bind_int(update_stmt, 3, arcEventInfo->update_count);
        if(SQLITE_OK != rc) {
            logit("et", "Error binding value in statment (%i): %s\n", rc, sqlite3_errmsg(db));
            return 4;
        }
        rc = sqlite3_bind_int(update_stmt, 4, arcEventInfo->flags);
        if(SQLITE_OK != rc) {
            logit("et", "Error binding value in statment (%i): %s\n", rc, sqlite3_errmsg(db));
            return 4;
        }
        rc = sqlite3_bind_text(update_stmt, 5, arcEventInfo->eventID, strlen(arcEventInfo->eventID), NULL);
        if(SQLITE_OK != rc) {
            logit("et", "Error binding value in statment (%i): %s\n", rc, sqlite3_errmsg(db));
            return 5;
        }
        rc = sqlite3_step(update_stmt);
        if(SQLITE_DONE != rc) {
            logit("et", "update statement didn't return DONE (%i): %s\n", rc, sqlite3_errmsg(db));
            logit("t", "Update event: '%s' -> '%010ld' (%f, %ld)\n", arcEventInfo->eventID, arcEventInfo->rowid, arcEventInfo->mag, arcEventInfo->version );
            return 6;
        }
        prefix = "Update";
    } else {
        rc = sqlite3_bind_text(insert_stmt, 1, arcEventInfo->eventID, strlen(arcEventInfo->eventID), NULL);
        if(SQLITE_OK != rc) {
            logit("et", "Error binding value in statment (%i): %s\n", rc, sqlite3_errmsg(db));
            return 7;
        }
        rc = sqlite3_bind_double(insert_stmt, 2, arcEventInfo->mag);
        if(SQLITE_OK != rc) {
            logit("et", "Error binding value in statment (%i): %s\n", rc, sqlite3_errmsg(db));
            return 8;
        }
        rc = sqlite3_bind_int64(insert_stmt, 3, arcEventInfo->version);
        if(SQLITE_OK != rc) {
            logit("et", "Error binding value in statment (%i): %s\n", rc, sqlite3_errmsg(db));
            return 9;
        }
        rc = sqlite3_bind_int(insert_stmt, 4, arcEventInfo->update_count);
        if(SQLITE_OK != rc) {
            logit("et", "Error binding value in statment (%i): %s\n", rc, sqlite3_errmsg(db));
            return 10;
        }
        rc = sqlite3_bind_int(insert_stmt, 5, arcEventInfo->flags);
        if(SQLITE_OK != rc) {
            logit("et", "Error binding value in statment (%i): %s\n", rc, sqlite3_errmsg(db));
            return 10;
        }
        rc = sqlite3_step(insert_stmt);
        if(SQLITE_DONE != rc) {
            logit("et", "insert statement didn't return DONE (%i): %s\n", rc, sqlite3_errmsg(db));
            return 11;
        }
        arcEventInfo->rowid = sqlite3_last_insert_rowid(db);
        prefix = "New";
    }
    if (arcEventInfo->update_count > 0) {
        logit("t", "%s event: '%s' v%ld UPDATE %d -> '%010ld' (%f)\n", prefix, arcEventInfo->eventID, arcEventInfo->version,
              arcEventInfo->update_count, arcEventInfo->rowid, arcEventInfo->mag );
    } else {
        logit("t", "%s event: '%s' v%ld -> '%010ld' (%f)\n", prefix, arcEventInfo->eventID, arcEventInfo->version,
              arcEventInfo->rowid, arcEventInfo->mag );
    }
    return 0;
}

/**
* Get the event information from  the ARC database.
* @param arcEventInfo the ARC Event Information
* Input is either eventID (if open for read only) or rowid
* Output is all
* @return 0 if success or non-zero  if error
*/
int arcdb_get_event(ARCEVENTINFO *arcEventInfo) {
    int rc;
    if ((rc = open_if_needed()) != SQLITE_OK) {
        return 1;
    }
    rc = do_get_event(arcEventInfo);
    reset_statement(select_stmt);
    if (rc != 0) {
        arcdb_close();
        rc = arcdb_open();
        if (rc == SQLITE_OK) {
            rc = do_get_event(arcEventInfo);
            reset_statement(select_stmt);
        } else {
            rc = 1;
        }
    }
    return rc;
}

/**
* Save the event to the ARC database.
* @param arcEventInfo the ARC Event Information
* @return 0 if success or non-zero  if error
*/
int arcdb_save_event(ARCEVENTINFO *arcEventInfo) {
    int rc = do_save_event(arcEventInfo);
    init_statements();
    if (rc != 0) {
        arcdb_close();
        rc = arcdb_open();
        if (rc == SQLITE_OK) {
            rc = do_save_event(arcEventInfo);
            init_statements();
        } else {
            rc = 1;
        }
    }
    return rc;
}
