/***************************************************************************
 *
 * A MSEED process function for Earthworm. Processes MSEED data records
 * and calls the specified callback handlers.
 *
 * Originally mseed2ewring in slink2ew which was written by Chad Trabant.
 ***************************************************************************/

#include "mseedprocess.h"
#include "earthworm.h"
#include "kom.h"
#include "libmseed.h"
#include "trace_buf.h"
#include "transport.h"
#include <string.h>

// Parse network, station, location and channel from an FDSN Source
// sid - source identifier
// nslc - the network, station, location and channel
// nslc_len - the length of the network, station, location and channel or NULL if none
// return 0 on success
int
ms_sid2nslc_t (const char *sid, ms_nslc_t *nslc, ms_nslc_len *nslc_len)
{
  int retval;
  size_t len;
  memset (nslc, 0, sizeof (ms_nslc_t));
  if (nslc_len)
  {
    memset (nslc_len, 0, sizeof (ms_nslc_len));
  }
  retval = ms_sid2nslc (sid, nslc->net, nslc->sta, nslc->loc, nslc->chan);
  if (!retval && nslc_len) // if success and length specfied
  {
    if ((len = strlen (nslc->net) + 1) > TRACE2_NET_LEN)
    {
      logit ("et", "mseedprocess: Error - network code too long for TRACE2_HEADER: %s\n", nslc->net);
      return 1;
    }
    nslc_len->netlen = (int)len;
    if ((len = strlen (nslc->sta) + 1) > TRACE2_STA_LEN)
    {
      logit ("et", "mseedprocess: Error - station code too long for TRACE2_HEADER: %s\n", nslc->sta);
      return 2;
    }
    nslc_len->stalen = (int)len;
    if ((len = strlen (nslc->loc) + 1) > TRACE2_LOC_LEN)
    {
      logit ("et", "mseedprocess: Error - location code too long for TRACE2_HEADER: %s\n", nslc->loc);
      return 3;
    }
    if (len <= 1) // if just null character
    {
      len = strlen (LOC_NULL_STRING) + 1;
      memcpy (nslc->loc, LOC_NULL_STRING, len);
    }
    nslc_len->loclen = (int)len;
    if ((len = strlen (nslc->chan) + 1) > TRACE2_CHAN_LEN)
    {
      logit ("et", "mseedprocess: Error - channel code too long for TRACE2_HEADER: %s\n", nslc->chan);
      return 4;
    }
    nslc_len->chanlen = (int)len;
  }
  return retval;
}

// Parse network, station, location and channel from an FDSN Source
// sid - source identifier
// trh2 - TRACEBUF2 header
// return 0 on success
int
ms_sid2tbuf (const char *sid, TRACE2_HEADER *trh2)
{
  ms_nslc_t nslc;
  ms_nslc_len nslc_len;
  /* Derive NSLC codes from Source ID and verify their lengths will fit */
  int retval = ms_sid2nslc_t (sid, &nslc, &nslc_len);
  if (retval)
  {
    return retval;
  }
  memcpy (trh2->net, nslc.net, nslc_len.netlen);
  memcpy (trh2->sta, nslc.sta, nslc_len.stalen);
  memcpy (trh2->loc, nslc.loc, nslc_len.loclen);
  memcpy (trh2->chan, nslc.chan, nslc_len.chanlen);
  return 0;
}

// Process a MSEED record and call the specified callback handlers
// msg - the message data
// msglength - the message length
// maxLatency - the maximum latency or 0 if none
// mh - the MSEED handler or NULL if none
// th - the TBUF handler or NULL if none
// returns 0 if successful
int
mseedprocess (char *msg, int msglength, double maxLatency,
              msg_handler mh, msg_handler th)
{
  static MS3Record *msr = NULL;
  int64_t reclen;
  uint8_t formatversion;
  uint32_t flags = MSF_UNPACKDATA;
  uint8_t samplesize;
  size_t maxsamplebytes;
  size_t samplebytes;
  uint8_t *dataptr = NULL;
  int maxsamples;
  void *samples;
  double now, latency;
  time_t t;

  /* Sanity check for MSEED record */
  reclen = ms3_detect (msg, msglength, &formatversion);
  if (reclen < 0)
  {
    logit ("et", "mseedprocess: Error - skipping bogus MSEED record.\n");
    return 1;
  }

  // if MSEED handler was specified
  if (mh != NULL)
  {
    /* call callback handler with MSEED record */
    if (mh (msg, msglength, formatversion))
    {
      logit ("et", "mseedprocess: Error handling MSEED message.\n");
      return 2;
    }
  }
  // if TBUF handler was specified
  if (th != NULL)
  {
    static TracePacket tbuf;

    /* Convert MSEED record to tracebuf2 and call callback handler */
    t   = time (NULL);
    now = (double)t;
    memset (&tbuf, 0, sizeof (TRACE2_HEADER));

    /* Parse MSEED record */
    if (msr3_parse (msg, msglength, &msr, flags, 0))
    {
      logit ("et", "mseedprocess: Error - skipping bogus MSEED record.\n");
      return 3;
    }
    /* Only records that contain data samples can be converted to TRACBUF2s */
    else if (msr->numsamples <= 0)
    {
      logit ("et", "mseedprocess: skipping MSEED record. numsamples <= 0\n");
      return 4;
    }
    else if (msr->samprate == 0.0)
    {
      logit ("et", "mseedprocess: skipping MSEED record. samprate == 0.0\n");
      return 5;
    }

    /* Derive NSLC codes from Source ID and verify their lengths will fit */
    if (ms_sid2tbuf (msr->sid, &tbuf.trh2))
    {
      logit ("et", "mseedprocess: Error deriving NSLC codes from: %s'.\n", msr->sid);
      return 6;
    }
    if (msr->numsamples > INT32_MAX)
    {
      logit ("et", "mseedprocess: Error - too many samples for TRACE2_HEADER: %" PRId64 "\n", msr->numsamples);
      return 7;
    }

    latency = now - tbuf.trh2.starttime;

    if (!maxLatency || (latency < maxLatency))
    {
      /* Create a TRACEBUF2 message */
      tbuf.trh2.pinno      = 0;
      tbuf.trh2.samprate   = msr3_sampratehz (msr);
      tbuf.trh2.version[0] = TRACE2_VERSION0;
      tbuf.trh2.version[1] = TRACE2_VERSION1;
      tbuf.trh2.quality[0] = 0;
      tbuf.trh2.quality[1] = 0;

      /* The decoding samples are in host byte order */
#ifdef _INTEL
      if (msr->sampletype == 'i')
        strcpy (tbuf.trh2.datatype, "i4");
      else if (msr->sampletype == 'f')
        strcpy (tbuf.trh2.datatype, "f4");
      else if (msr->sampletype == 'd')
        strcpy (tbuf.trh2.datatype, "f8");
      else
      {
        logit ("et", "mseedprocess: Error - unrecognized sample type '%c'\n",
               msr->sampletype);
        return 8;
      }
#endif
#ifdef _SPARC
      if (msr->sampletype == 'i')
        strcpy (tbuf.trh2.datatype, "s4");
      else if (msr->sampletype == 'f')
        strcpy (tbuf.trh2.datatype, "t4");
      else if (msr->sampletype == 'd')
        strcpy (tbuf.trh2.datatype, "t8");
      else
      {
        logit ("et", "mseedprocess: Error - unrecognized sample type '%c'\n",
               msr->sampletype);
        return 9;
      }
#endif

      /* Determine the maximum number of samples in a tracebuf2 */
      samplesize     = ms_samplesize (msr->sampletype);
      maxsamplebytes = MAX_TRACEBUF_SIZ - sizeof (TRACE2_HEADER);
      maxsamples     = (int)(maxsamplebytes / samplesize);

      /* Create and send as many tracebufs2 as needed for the data */
      samples = (void *)((char *)&tbuf + sizeof (TRACE2_HEADER));
      dataptr = msr->datasamples;

      while (msr->numsamples)
      {
        tbuf.trh2.nsamp     = (msr->numsamples > maxsamples) ? maxsamples : (int)msr->numsamples;
        tbuf.trh2.starttime = (double)MS_NSTIME2EPOCH (msr->starttime);
        tbuf.trh2.endtime   = (double)MS_NSTIME2EPOCH (ms_sampletime (msr->starttime, tbuf.trh2.nsamp - 1, tbuf.trh2.samprate));

        samplebytes = tbuf.trh2.nsamp * samplesize;
        memcpy (samples, dataptr, samplebytes);
        if (th ((char *)&tbuf, samplebytes + sizeof (TRACE2_HEADER),
                MSG_TBUF))
        {
          logit ("et", "mseedprocess: Error handling TBUF message.\n");
          return 10;
        }

        /* Adjust values to account for what was just transmitted */
        msr->numsamples -= tbuf.trh2.nsamp;
        msr->starttime = ms_sampletime (msr->starttime, tbuf.trh2.nsamp, tbuf.trh2.samprate);
        dataptr += samplebytes;
      }
    }
    else
    {
      logit ("et", "mseedprocess: Latent Packet for %5s.%3s.%2s.%2s %lf secs\n",
             tbuf.trh2.sta, tbuf.trh2.chan, tbuf.trh2.net, tbuf.trh2.loc, latency);
    }
  }

  return 0;
}

// Unpack NSLC from a MSEED record
// msg - the message data
// msglength - the message length
// nslc - the network, station, location and channel
// returns 0 if successful
int
ms_unpacknslc (char *msg, int msglength, ms_nslc_t *nslc)
{
  static MS3Record *msr = NULL;
  uint32_t flags        = MSF_UNPACKDATA;
  int retval;

  /* Parse MSEED record */
  if (msr3_parse (msg, msglength, &msr, flags, 0))
  {
    logit ("et", "ms_unpacksid: could not parse MSEED record.\n");
    return -2;
  }
  // Parse network, station, location and channel from the FDSN Source
  retval = ms_sid2nslc_t (msr->sid, nslc, NULL);
  if (retval)
  {
    logit ("et", "ms_unpacksid: could not parse NSLC.\n");
    return retval;
  }
  return 0;
}
