/*
 *   THIS FILE IS UNDER RCS - DO NOT MODIFY UNLESS YOU HAVE
 *   CHECKED IT OUT USING THE COMMAND CHECKOUT.
 *
 *    $Id: readcnfg.c 7390 2018-05-02 01:55:51Z baker $
 *
 *    Revision history:
 *     $Log$
 *     Revision 1.8  2009/08/28 17:48:36  paulf
 *     added TrigIdFilename option for pointing to trig_id.d
 *
 *     Revision 1.7  2004/05/11 17:49:07  lombard
 *     Added support for location code, TYPE_CARLSTATRIG_SCNL and TYPE_TRIGLIST_SCNL
 *     messages.
 *     Removed OS2 support
 *
 *     Revision 1.6  2002/06/05 14:52:46  patton
 *     Made logit changes.
 *
 *     Revision 1.5  2000/11/01 22:35:05  dietz
 *     csuNet->eventID is now read with k_long instead of k_val.
 *
 *     Revision 1.4  2000/08/08 18:33:24  lucky
 *     Lint cleanup
 *
 *     Revision 1.3  2000/06/10 16:25:22  lombard
 *     clean up return conditions
 *
 *     Revision 1.2  2000/06/10 04:15:47  lombard
 *     Fixed bug that caused crash on zero-length lines in config files.
 *
 *     Revision 1.1  2000/02/14 16:14:42  lucky
 *     Initial revision
 *
 *
 */

/*
 * readcnfg.c: Read the CarlSubTrig parameters from a file.
 *              1) Set members of the NETWORK and CTPARAM structures.
 */

/*******                                                        *********/
/*      Functions defined in this source file                           */
/*******                                                        *********/

/*      Function: ReadConfig                                            */
/*                                                                      */
/*      Inputs:         Pointer to a string(input filename)             */
/*                      Pointer to the CarlSubTrig Network structure    */
/*                                                                      */
/*      Outputs:        Updated CarlSubTrig structures (above)  */
/*                      Error messages to stderr                        */
/*                                                                      */
/*      Returns:        0 on success                                    */
/*                      Error code as defined in carlsubtrig.h on       */
/*                        failure                                       */

/*******                                                        *********/
/*      System Includes                                                 */
/*******                                                        *********/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>     /* strcpy                                       */

/*******                                                        *********/
/*      Earthworm Includes                                              */
/*******                                                        *********/
#include "earthworm.h"  /* GetKey                                       */
#include "kom.h"        /* k_close, k_err, k_int, k_its, k_open, k_rd,  */
#include "k_readcfg.h"
/*   k_str, k_val                               */

#define MAXBUF  4       /* found in kom.c ( should move to kom.h ??? )  */
#define NUMCTPARAMS     25      /* Number of parameters in CSUPARAM that*/
/*   can be set from the config file.   */
#define NUMCTREQ        12      /* Number of parameters in CSUPARAM that*/
/*   MUST be set from the config file.  */

/*******                                                        *********/
/*      CarlSubTrig Includes                                            */
/*******                                                        *********/
#include "carlsubtrig.h"

/*******                                                        *********/
/*      Functions referenced in this source file                        */
/*******                                                        *********/

/*******                                                        *********/
/*      Function definitions                                            */
/*******                                                        *********/

typedef struct {
  NETWORK* csuNet;
  char  params[NUMCTPARAMS];    /* Flag for each parameter that is set. */
  char* paramNames[NUMCTPARAMS];        /* Name of each parameter.      */
} PROCESS_DATA;

/*      Function: processCommand                                            */
static int processCommand(READCFG_DATA *data)
{
  PROCESS_DATA* p = (PROCESS_DATA*)data->commandData;
  NETWORK* csuNet = p->csuNet;
  int   result;
  char* value;

        /*      Look for a configuration parameter - required first     */
        if ( k_its( p->paramNames[0] ) )
        {
          /*    Read the MyModuleId value                               */
          value = k_str( );
          if ( value != NULL )
          {
            strcpy( csuNet->csuParam.myModName, value );
            p->params[0] = 1;
          }
        }
        else if ( k_its( p->paramNames[1] ) )
        {
          /*    Read the RingNameIn value                               */
          value = k_str( );
          if ( value != NULL )
          {
            strcpy( csuNet->csuParam.ringIn, value );
            if ( -1 == ( csuNet->csuParam.ringInKey = 
                         GetKey( csuNet->csuParam.ringIn ) ) )
            {
              logit( "e", "CarlSubTrig: Error finding key for ring '%s'.\n",
                       csuNet->csuParam.ringIn );
              return (ERR_CONFIG_READ);
            }
            else
              p->params[1] = 1;
          }
        }
        else if ( k_its( p->paramNames[2] ) )
        {
          /*    Read the RingNameOut value                              */
          value = k_str( );
          if ( value != NULL )
          {
            strcpy( csuNet->csuParam.ringOut, value );
            if ( -1 == ( csuNet->csuParam.ringOutKey = 
                         GetKey( csuNet->csuParam.ringOut ) ) )
            {
              logit( "e", "CarlSubTrig: Error finding key for ring '%s'.\n",
                       csuNet->csuParam.ringOut );
              return (ERR_CONFIG_READ);
            }
            else
              p->params[2] = 1;
          }
        }
        else if ( k_its( p->paramNames[3] ) )
        {
          /*    Read the HeartBeatInterval value                        */
          csuNet->csuParam.heartbeatInt = k_int( );
          p->params[3] = 1;
        }
        else if ( k_its( p->paramNames[4] ) )
        {
          /*    Read the GetEventsFrom values                           */
          value = k_str( );
          if ( value != NULL )
          {
            strcpy( csuNet->csuParam.readInstName, value );
            value = k_str( );
            if ( value != NULL )
            {
              strcpy( csuNet->csuParam.readModName, value );
              p->params[4] = 1;
            }
          }
        }
        else if ( k_its( p->paramNames[5] ) )
        {
          /*    Read the StationFile value                              */
          value = k_str( );
          if ( value != NULL )
          {
            strcpy( csuNet->csuParam.staFile, value );
            p->params[5] = 1;
          }
        }
        else if ( k_its( p->paramNames[6] ) )
        {
          /*    Read the Latency value                                  */
          csuNet->latency =  k_long( );
          p->params[6] = 1;

          /* During the latency period, we can receive at most one      */
          /*   trigger message per second for each station. Assuming    */
          /*   half of these are ON and half of these are OFF messages, */
          /*   we need latency / 2 slots for pending triggers, and a    */
          /*   minimum of 4 to give some operating room.                */
          csuNet->nSlots = csuNet->latency / 2;
          if ( csuNet->nSlots < 4 ) csuNet->nSlots = 4;
        }
        else if ( k_its( p->paramNames[7] ) )
        {
          /*    Read the Base Network Trigger Duration                  */
          csuNet->NetTrigDur = k_long( );
          p->params[7] = 1;
        }
        else if ( k_its( p->paramNames[8] ) )
        {
          /*    Read the Subnet Contribution value                      */
          csuNet->subnetContrib = k_long( );
          p->params[8] = 1;
        }
        else if ( k_its( p->paramNames[9] ) )
        {
          /*    Read the PreEventTime value                             */
          csuNet->PreEventTime = k_long( );
          p->params[9] = 1;
        }
        else if ( k_its( p->paramNames[10] ) )
        {
          /*    Read the Maximum Duration                               */
          csuNet->MaxDur = k_long( );
          p->params[10] = 1;
        }
        else if ( k_its( p->paramNames[11] ) )
        {
          /*    Read the Default Station Duration                       */
          csuNet->DefaultStationDur = k_long( );
          p->params[11] = 1;
        }
        else if ( k_its( p->paramNames[12] ) )
        {
          /*    Read the next event ID number                           */
          csuNet->eventID = k_long( );
          p->params[12] = 1;
        }
        else if ( k_its( p->paramNames[13] ) )
        {
          /*    Read the Debug value                                    */
          csuNet->csuParam.debug = k_int( );
          p->params[13] = 1;
        }
        else if ( k_its( p->paramNames[14] ) )
        {
          csuNet->nSub++;
        }
        else if ( k_its( p->paramNames[15] ) )
        {
          value = k_str( );
          if ( value != NULL )
          {
            char  staCode[TRACE2_STA_LEN];
            char  compCode[TRACE2_CHAN_LEN];
            char  netCode[TRACE2_NET_LEN];
            char  locCode[TRACE2_LOC_LEN];
            if ( sscanf( value, "%[^.].%[^.].%[^.].%s", staCode, compCode, 
                         netCode, locCode ) == 4 )
            {
              csuNet->channels =
                   (CHANNEL *) realloc(csuNet->channels,
                                       ( csuNet->nChan + 1 ) * sizeof( CHANNEL ) );
              if ( csuNet->channels != NULL )
              {
                strncpy( csuNet->channels[csuNet->nChan].staCode, staCode, 
			 TRACE2_STA_LEN );
                strncpy( csuNet->channels[csuNet->nChan].compCode, compCode, 
			 TRACE2_CHAN_LEN );
                strncpy( csuNet->channels[csuNet->nChan].netCode, netCode, 
			 TRACE2_NET_LEN );
                strncpy( csuNet->channels[csuNet->nChan].locCode, locCode, 
			 TRACE2_LOC_LEN );
                csuNet->nChan++;
              } 
              else 
              {
                logit( "e", "carlsubtrig: Error allocating channel memory." );
                return (ERR_MALLOC);
              }
            }
            else 
            {       /*  Encountered a parsing problem           */
              logit( "e", "ReadConfig: Failed to parse channel: '%s'.\n", value );
              return (ERR_CONFIG_READ);
            }
            p->params[15] = 1;
          }
        }
        
        else if ( k_its( p->paramNames[16] ) )
        {
          /* Read the list-subnets value */
          csuNet->listSubnets = k_val( );
          p->params[16] = 1;
        }
        
        else if ( k_its( p->paramNames[17] ) )
        {
          /* Read the numSubAll value */
          csuNet->numSubAll = k_val( );
          p->params[17] = 1;
        }

        else if ( k_its( p->paramNames[18] ) )
        {
          /* Set the component as wildcard flag */
          csuNet->compAsWild = 1;
          p->params[18] = 1;
        }
        else if ( k_its( p->paramNames[19] ) )
        {
          /* Set maximum length of a triglist message */
          csuNet->trigMsgBufLen = k_long( );
          if( csuNet->trigMsgBufLen > MAX_BYTES_PER_EQ ) 
            csuNet->trigMsgBufLen = MAX_BYTES_PER_EQ;
          p->params[19] = 1;
        }
        else if ( k_its( p->paramNames[20] ) )
        {
          csuNet->csuParam.trigIdFilename = strdup(k_str( ));
          p->params[20] = 1;
          /* okay, now here is where it gets hairy, we need to open this file and parse the next_id token inside it */
          result = k_opencfg( data, csuNet->csuParam.trigIdFilename );
          if ( result == 0 )
          {
            logit( "e",
                     "CarlSubTrig: Error opening trig_id file '%s'.\n",
                     csuNet->csuParam.trigIdFilename );
            return (ERR_CONFIG_OPEN);
          }
        }
        else if ( k_its( p->paramNames[21] ) )  /* number of coincident stations to throw out a subnet trigger */
        {
          /* EXPERIMENTAL */
	    csuNet->coincident_stas=k_int();
        }
        else if ( k_its( p->paramNames[22] ) )  /* if more than this subnets trigger, ignore coincident check */
        {
          /* EXPERIMENTAL */
	    csuNet->ignore_coincident=k_int();
        }
        else if ( k_its( p->paramNames[23] ) )  /* if 1, launch early warning */
        {
          /* EXPERIMENTAL */
	    csuNet->early_warning=k_int();
        }
        else if ( k_its( p->paramNames[24] ) )  /* if 1, set to playback mode*/
        {
          /* EXPERIMENTAL */
	    csuNet->useDataTime=k_int();
        }
        /*      Unknown parameter found                                 */
        else
        {
          fprintf( stderr, "CarlSubTrig: Unknown parameter '%s' in '%s'.\n",
                   data->token, data->filesOpen[getNumOpen() - 1] );
          return -1;
        }
  return 0;
}

/*      Function: ReadConfig                                            */
int ReadConfig( char* filename, NETWORK* csuNet )
{
  PROCESS_DATA p;
  READCFG_DATA data;
  int   index;                  /* Loop control variable.               */
  int   result;

  p.csuNet = csuNet;
  k_initcfg(&data, &p);
  /*    Initialize the parameter names                                  */
  p.paramNames[0] = "MyModuleId";
  p.paramNames[1] = "RingNameIn";
  p.paramNames[2] = "RingNameOut";
  p.paramNames[3] = "HeartBeatInterval";
  p.paramNames[4] = "GetEventsFrom";
  p.paramNames[5] = "StationFile";
  p.paramNames[6] = "Latency";
  p.paramNames[7] = "NetTriggerDur";
  p.paramNames[8] = "SubnetContrib";
  p.paramNames[9] = "PreEventTime";
  p.paramNames[10] = "MaxDuration";
  p.paramNames[11] = "DefStationDur";
  p.paramNames[12] = "next_id";
  p.paramNames[13] = "Debug";
  p.paramNames[14] = SUBNET_PARAM;
  p.paramNames[15] = "Channel";
  p.paramNames[16] = "ListSubnets";
  p.paramNames[17] = "AllSubnets";
  p.paramNames[18] = "CompAsWild";
  p.paramNames[19] = "MaxTrigMsgLen";
  p.paramNames[20] = "TrigIdFilename";
  p.paramNames[21] = "CoincidentStaTriggers";
  p.paramNames[22] = "IgnoreCoincident";
  p.paramNames[23] = "EarlyWarning";   /* RSL: For early warning parameter */
  p.paramNames[24] = "PlaybackMode";   /* paulf for using data time, not system */

  csuNet->coincident_stas=0;
  csuNet->ignore_coincident=3;
  csuNet->early_warning = 0; /* RSL: Default setting does not send early warning */
  csuNet->useDataTime =0;
  
  /*    Initialize the parameter set flags                              */
  for ( index = 0; index < NUMCTPARAMS; index++ )
  {
    p.params[index] = 0;
  }
  /* default to the old trig_id.d */
  csuNet->csuParam.trigIdFilename = NULL;
  /* The subnet commands are in the main config file                    */
  strcpy( csuNet->csuParam.subFile, filename );

  result = k_readcfg(&data, filename, &processCommand);
  if (result == 0) {
    logit("", "CarlSubTrig: %d  subnets found in command file <%s>\n",
          csuNet->nSub, filename);
  }
  return result;
}
